/*
 * Copyright (c) 2022 Xilinx
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <stdio.h>
#include <stddef.h>
#include <time.h>
#if defined(__linux__)
#include <sys/time.h>
#elif defined(_WIN32)
#include <winsock.h>
#include <WinApi.h>
#endif
#include <float.h>
#include "vpi_api.h"
#include "vpi_types.h"

#include "codec_internal.h"
#include "internal.h"
#include "hwconfig.h"
#include "encode.h"
#include "libavutil/hwcontext_vpe.h"
#include "libavutil/time.h"
#include "libavutil/opt.h"
#include "libavutil/pixfmt.h"
#include "libavutil/pixdesc.h"
#include "libavutil/fifo.h"

#if CONFIG_LIBXRM
#include "xrm_enc_interface.h"
#endif
#include "ama_dynamic_params.h"

#include "log_ama.h"

typedef struct snEncContext {
    const AVClass    *class;
    AVBufferRef      *hwframe;
    AVBufferRef      *hwdevice;
    VpiEncHandle     enc_handle;
    VpiEncInitParams enc_params;
    char            *dynamic_param_location;
    FILE            *dynamic_params_fp;

#if CONFIG_LIBXRM
    XrmEncodeContext xrm_ctx;
#endif

    AVFrame *frame;
    uint32_t eof_flush_en;
    uint32_t flush_done;
    uint32_t sendFrameNum;
    uint32_t recvFrameNum;

    /* add for perf test */
    int64_t  start_time;
    int64_t  end_time;
    /* add end */

    VpiPkt *vpi_pkt;
    int initialized;

    unsigned int dts_delay;
    AVFifo *pts_queue;

    int nb_components;
    double sum_psnr[3];
    double sum_ssim[3];
    double max_psnr[3];
    double max_ssim[3];
    double min_psnr[3];
    double min_ssim[3];

    char instance_name[64];
} snEncContext;

/* command line options */
#define OFFSETENCPARAM(x)   offsetof(snEncContext, enc_params) + offsetof(VpiEncInitParams, x)
#define OFFSET(x)           offsetof(snEncContext, x)
#define FLAGS               (AV_OPT_FLAG_VIDEO_PARAM | AV_OPT_FLAG_ENCODING_PARAM | AV_OPT_FLAG_EXPORT)

static int sn_encode_start(AVCodecContext *avctx);

static const AVOption h264enc_options[] = {
    { "level", "Set the encoding level restriction", OFFSETENCPARAM(enc_level), AV_OPT_TYPE_INT, {.i64 = 0}, 0, 62, FLAGS, "level" },
        { "auto", "0 Auto",        0, AV_OPT_TYPE_CONST, {.i64 = 0},  0, 0, FLAGS, "level"},
        { "1",    "1 level(10)",   0, AV_OPT_TYPE_CONST, {.i64 = 10}, 0, 0, FLAGS, "level"},
        { "1.1",  "1.1 level(11)", 0, AV_OPT_TYPE_CONST, {.i64 = 11}, 0, 0, FLAGS, "level"},
        { "1.2",  "1.2 level(12)", 0, AV_OPT_TYPE_CONST, {.i64 = 12}, 0, 0, FLAGS, "level"},
        { "1.3",  "1.3 level(13)", 0, AV_OPT_TYPE_CONST, {.i64 = 13}, 0, 0, FLAGS, "level"},
        { "2",    "2 level(20)",   0, AV_OPT_TYPE_CONST, {.i64 = 20}, 0, 0, FLAGS, "level"},
        { "2.1",  "2.1 level(21)", 0, AV_OPT_TYPE_CONST, {.i64 = 21}, 0, 0, FLAGS, "level"},
        { "2.2",  "2.2 level(22)", 0, AV_OPT_TYPE_CONST, {.i64 = 22}, 0, 0, FLAGS, "level"},
        { "3",    "3 level(30)",   0, AV_OPT_TYPE_CONST, {.i64 = 30}, 0, 0, FLAGS, "level"},
        { "3.1",  "3.1 level(31)", 0, AV_OPT_TYPE_CONST, {.i64 = 31}, 0, 0, FLAGS, "level"},
        { "3.2",  "3.2 level(32)", 0, AV_OPT_TYPE_CONST, {.i64 = 32}, 0, 0, FLAGS, "level"},
        { "4",    "4 level(40)",   0, AV_OPT_TYPE_CONST, {.i64 = 40}, 0, 0, FLAGS, "level"},
        { "4.1",  "4.1 level(41)", 0, AV_OPT_TYPE_CONST, {.i64 = 41}, 0, 0, FLAGS, "level"},
        { "4.2",  "4.2 level(42)", 0, AV_OPT_TYPE_CONST, {.i64 = 42}, 0, 0, FLAGS, "level"},
        { "5",    "5 level(50)",   0, AV_OPT_TYPE_CONST, {.i64 = 50}, 0, 0, FLAGS, "level"},
        { "5.1",  "5.1 level(51)", 0, AV_OPT_TYPE_CONST, {.i64 = 51}, 0, 0, FLAGS, "level"},
        { "5.2",  "5.2 level(52)", 0, AV_OPT_TYPE_CONST, {.i64 = 52}, 0, 0, FLAGS, "level"},
        { "6",    "6 level(60)",   0, AV_OPT_TYPE_CONST, {.i64 = 60}, 0, 0, FLAGS, "level"},
        { "6.1",  "6.1 level(61)", 0, AV_OPT_TYPE_CONST, {.i64 = 61}, 0, 0, FLAGS, "level"},
        { "6.2",  "6.2 level(62)", 0, AV_OPT_TYPE_CONST, {.i64 = 62}, 0, 0, FLAGS, "level"},

    { "profile", "Set the encoding profile", OFFSETENCPARAM(enc_profile), AV_OPT_TYPE_INT, {.i64 = PROFILE_AUTO}, -1, PROFILE_H264_HIGH10_INTRA, FLAGS, "profile"},
        { "auto",         "", 0,  AV_OPT_TYPE_CONST, {.i64 = PROFILE_AUTO},              0, 0, FLAGS, "profile"},
        { "baseline",     "", 0,  AV_OPT_TYPE_CONST, {.i64 = PROFILE_H264_BASELINE},     0, 0, FLAGS, "profile"},
        { "main",         "", 0,  AV_OPT_TYPE_CONST, {.i64 = PROFILE_H264_MAIN},         0, 0, FLAGS, "profile"},
        { "high",         "", 0,  AV_OPT_TYPE_CONST, {.i64 = PROFILE_H264_HIGH},         0, 0, FLAGS, "profile"},
        { "high-10",       "", 0,  AV_OPT_TYPE_CONST, {.i64 = PROFILE_H264_HIGH10},       0, 0, FLAGS, "profile"},
        { "high-10-intra", "", 0,  AV_OPT_TYPE_CONST, {.i64 = PROFILE_H264_HIGH10_INTRA}, 0, 0, FLAGS, "profile"},

    { "control_rate", "Rate control mode", OFFSETENCPARAM(rate_control), AV_OPT_TYPE_INT, {.i64 = MODE_RC_AUTO}, -1, 4, FLAGS, "control_rate"},
        { "auto",     "auto(-1)",                         0,  AV_OPT_TYPE_CONST, {.i64 = MODE_RC_AUTO},     0, 0, FLAGS, "control_rate"},
        { "const_qp", "contant qp(0)",                    0,  AV_OPT_TYPE_CONST, {.i64 = MODE_RC_CONST_QP}, 0, 0, FLAGS, "control_rate"},
        { "cbr",      "constabt bit-rate(1)",             0,  AV_OPT_TYPE_CONST, {.i64 = MODE_RC_CBR},      0, 0, FLAGS, "control_rate"},
        { "vbr",      "variable bit-rate(2)",             0,  AV_OPT_TYPE_CONST, {.i64 = MODE_RC_VBR},      0, 0, FLAGS, "control_rate"},
        { "cvbr",     "constrained variable bit-rate(3)", 0,  AV_OPT_TYPE_CONST, {.i64 = MODE_RC_CVBR},     0, 0, FLAGS, "control_rate"},
        { "cabr",     "content adaptive bit-rate(4)",     0,  AV_OPT_TYPE_CONST, {.i64 = MODE_RC_CABR},     0, 0, FLAGS, "control_rate"},

    { "preset", "Encoder preset Possible values: (slow, medium, fast)", OFFSETENCPARAM(preset), AV_OPT_TYPE_STRING, {.str = "medium"}, .flags=FLAGS, .unit="preset"},

    { "bf", "Number of B-frames", OFFSETENCPARAM(num_b_frames), AV_OPT_TYPE_INT, { .i64 = -1}, -1, 3, FLAGS, "b_frames"},
    { "qp", "Quantization parameter", OFFSETENCPARAM(qp), AV_OPT_TYPE_INT, {.i64 = -1}, -1, 51, FLAGS, "qp"},
    { "min_bitrate", "Minimum bit rate", OFFSETENCPARAM(min_bitrate), AV_OPT_TYPE_INT64, {.i64 = -1}, -1, 35000000000, FLAGS, "min_bitrate"},
    { "max_bitrate", "Maximum bit rate", OFFSETENCPARAM(max_bitrate), AV_OPT_TYPE_INT64, {.i64 = -1}, -1, 35000000000, FLAGS, "max_bitrate"},
    { "slice", "Slice num to use for encoding", OFFSETENCPARAM(slice_id), AV_OPT_TYPE_INT, { .i64 = -1}, -1, 1, FLAGS, "slice"},
    { "cores", "Cores decide if it's 1 or 2-slice encoding", OFFSETENCPARAM(num_cores), AV_OPT_TYPE_INT, { .i64 = 1}, 1, 1, FLAGS, "cores"},

    { "crf", "Enable/Disable CRF mode", OFFSETENCPARAM(crf),  AV_OPT_TYPE_INT, {.i64 = 0},  0, 1, FLAGS, "crf"},
        { "disable", "", 0,  AV_OPT_TYPE_CONST, {.i64 = 0}, 0, 0, FLAGS, "crf"},
        { "enable", "",  0,  AV_OPT_TYPE_CONST, {.i64 = 1}, 0, 0, FLAGS, "crf"},

    { "forced_idr", "If forcing keyframes, force them as IDR frames.", OFFSETENCPARAM(enc_force_idr),  AV_OPT_TYPE_INT, {.i64 = 1},  0, 1, FLAGS, "forced_idr"},
        { "disable", "", 0,  AV_OPT_TYPE_CONST, {.i64 = 0}, 0, 0, FLAGS, "forced_idr"},
        { "enable", "",  0,  AV_OPT_TYPE_CONST, {.i64 = 1}, 0, 0, FLAGS, "forced_idr"},

    { "lookahead_depth", "Number of frames in pipeline prior to the encoder", OFFSETENCPARAM(la_depth), AV_OPT_TYPE_INT, {.i64 = -1},  -1, 53, FLAGS, "lookahead_depth"},
    { "latency_ms", "Lookahead depth specified in milliseconds", OFFSETENCPARAM(latency_ms), AV_OPT_TYPE_INT, {.i64 = -1},  -1, 60000, FLAGS, "latency_ms"},
    { "no_bll", "No bframes for low latency usecase", OFFSETENCPARAM(no_low_latency_b_frames), AV_OPT_TYPE_INT, {.i64 = -1},  -1, 1, FLAGS, "no_bll"},

    { "qp_mode", "Quantization parameter", OFFSETENCPARAM(qp_mode), AV_OPT_TYPE_INT, { .i64 = MODE_QP_AUTO}, 0, 2, FLAGS, "qp_mode"},
        { "auto",          "best VQ",                  0, AV_OPT_TYPE_CONST, {.i64 = MODE_QP_AUTO},     0, 0, FLAGS, "qp_mode"},
        { "relative_load", "spatial AQ + temporal AQ", 0, AV_OPT_TYPE_CONST, {.i64 = MODE_QP_REL_LOAD}, 0, 0, FLAGS, "qp_mode"},
        { "uniform",        "AQ off",                  0, AV_OPT_TYPE_CONST, {.i64 = MODE_QP_UNIFORM},  0, 0, FLAGS, "qp_mode"},

    { "bufsize", "Size of VBV buffer (in bits)", OFFSETENCPARAM(bufsize), AV_OPT_TYPE_INT, { .i64 = -1}, -1, 400000000, FLAGS, "bufsize"},

    { "spatial_aq", "Enable spatial AQ", OFFSETENCPARAM(spatial_aq), AV_OPT_TYPE_INT, {.i64 = -1}, -1, 1, FLAGS, "spatial_aq"},
        { "auto",    "", 0,  AV_OPT_TYPE_CONST, {.i64 = -1}, 0, 0, FLAGS, "spatial_aq"},
        { "disable", "", 0,  AV_OPT_TYPE_CONST, {.i64 = 0},  0, 0, FLAGS, "spatial_aq"},
        { "enable", "",  0,  AV_OPT_TYPE_CONST, {.i64 = 1},  0, 0, FLAGS, "spatial_aq"},

    { "temporal_aq", "Enable temporal AQ", OFFSETENCPARAM(temporal_aq), AV_OPT_TYPE_INT, {.i64 = -1}, -1, 1, FLAGS, "temporal_aq"},
        { "auto",    "", 0,  AV_OPT_TYPE_CONST, {.i64 = -1}, 0, 0, FLAGS, "temporal_aq"},
        { "disable", "", 0,  AV_OPT_TYPE_CONST, {.i64 = 0},  0, 0, FLAGS, "temporal_aq"},
        { "enable",  "", 0,  AV_OPT_TYPE_CONST, {.i64 = 1},  0, 0, FLAGS, "temporal_aq"},

    { "spatial_aq_gain",  "Spatial aq gain scale factor",  OFFSETENCPARAM(spatial_aq_gain),  AV_OPT_TYPE_INT, {.i64 = 255}, 0, 255, FLAGS, "spatial_aq_gain"},
        { "auto", "", 0,  AV_OPT_TYPE_CONST, {.i64 = 255}, 0, 0, FLAGS, "spatial_aq_gain"},
    { "temporal_aq_gain", "Temporal aq gain scale factor", OFFSETENCPARAM(temporal_aq_gain), AV_OPT_TYPE_INT, {.i64 = 255}, 0, 255, FLAGS, "temporal_aq_gain"},
        { "auto", "", 0,  AV_OPT_TYPE_CONST, {.i64 = 255}, 0, 0, FLAGS, "temporal_aq_gain"},

    { "min_qp", "Min Qp for rate control", OFFSETENCPARAM(min_qp), AV_OPT_TYPE_INT, {.i64 = -1}, -1, 51, FLAGS, "min_qp"},
    { "max_qp", "Min Qp for rate control", OFFSETENCPARAM(max_qp), AV_OPT_TYPE_INT, {.i64 = -1}, -1, 51, FLAGS, "max_qp"},
    { "qp_i_offset", "QP Offset - I", OFFSETENCPARAM(qp_i_offset), AV_OPT_TYPE_INT, {.i64 = 0}, -51, 51, FLAGS, "qp_i_offset"},
    { "qp_b_offset", "QP Offset - B/P", OFFSETENCPARAM(qp_b_offset), AV_OPT_TYPE_INT, {.i64 = 0}, -51, 51, FLAGS, "qp_b_offset"},

    { "tune_metrics", "Tune VQ for objective metrics", OFFSETENCPARAM(tune_metrics), AV_OPT_TYPE_INT, {.i64 = TM_BEST_VQ}, 1, 4, FLAGS, "tune_metrics"},
        { "vq",      "best VQ (1)",   0, AV_OPT_TYPE_CONST, {.i64 = TM_BEST_VQ},   0, 0, FLAGS, "tune_metrics"},
        { "psnr",    "best PSNR (2)", 0, AV_OPT_TYPE_CONST, {.i64 = TM_BEST_PSNR}, 0, 0, FLAGS, "tune_metrics"},
        { "ssim",    "best SSIM (3)", 0, AV_OPT_TYPE_CONST, {.i64 = TM_BEST_SSIM}, 0, 0, FLAGS, "tune_metrics"},
        { "vmaf",    "best VMAF (4)", 0, AV_OPT_TYPE_CONST, {.i64 = TM_BEST_VMAF}, 0, 0, FLAGS, "tune_metrics"},

    { "latency_logging", "Log latency info to syslog", OFFSETENCPARAM(latency_logging), AV_OPT_TYPE_INT, {.i64 = 0}, 0, 1, FLAGS, "latency_logging"},
        { "disable", "", 0,  AV_OPT_TYPE_CONST, {.i64 = 0}, 0, 0, FLAGS, "latency_logging"},
        { "enable", "",  0,  AV_OPT_TYPE_CONST, {.i64 = 1}, 0, 0, FLAGS, "latency_logging"},

    { "dynamic_gop", "Dynamically determine the optimal number of B frames, based on the content.", OFFSETENCPARAM(dynamic_gop),  AV_OPT_TYPE_INT, {.i64 = -1}, -1, 1, FLAGS, "dynamic_gop"},
        { "auto",    "", 0,  AV_OPT_TYPE_CONST, {.i64 = -1}, 0, 0, FLAGS, "dynamic_gop"},
        { "disable", "", 0,  AV_OPT_TYPE_CONST, {.i64 = 0},  0, 0, FLAGS, "dynamic_gop"},
        { "enable",  "", 0,  AV_OPT_TYPE_CONST, {.i64 = 1},  0, 0, FLAGS, "dynamic_gop"},

    { "dynamic_params_file", "Configuration file to dynamically change codec params", OFFSET(dynamic_param_location), AV_OPT_TYPE_STRING, {.str = NULL}, .flags=FLAGS, .unit="dynamic_params_file"},

    { "expert_options", "encoder configuration using a ',' separated list of key=value parameters", OFFSETENCPARAM(xav1_options), AV_OPT_TYPE_STRING, {.str = NULL}, .flags=FLAGS, .unit="expert_options"},
    { NULL },
};

static const AVOption hevcenc_options[] = {
    { "level", "Set the encoding level restriction", OFFSETENCPARAM(enc_level), AV_OPT_TYPE_INT, { .i64 = 0 }, 0, 62, FLAGS, "level" },
        { "auto", "0 Auto",        0, AV_OPT_TYPE_CONST, {.i64 = 0},  0, 0, FLAGS, "level"},
        { "1",    "1 level(10)",   0, AV_OPT_TYPE_CONST, {.i64 = 10}, 0, 0, FLAGS, "level"},
        { "2",    "2 level(20)",   0, AV_OPT_TYPE_CONST, {.i64 = 20}, 0, 0, FLAGS, "level"},
        { "2.1",  "2.1 level(21)", 0, AV_OPT_TYPE_CONST, {.i64 = 21}, 0, 0, FLAGS, "level"},
        { "3",    "3 level(30)",   0, AV_OPT_TYPE_CONST, {.i64 = 30}, 0, 0, FLAGS, "level"},
        { "3.1",  "3.1 level(31)", 0, AV_OPT_TYPE_CONST, {.i64 = 31}, 0, 0, FLAGS, "level"},
        { "4",    "4 level(40)",   0, AV_OPT_TYPE_CONST, {.i64 = 40}, 0, 0, FLAGS, "level"},
        { "4.1",  "4.1 level(41)", 0, AV_OPT_TYPE_CONST, {.i64 = 41}, 0, 0, FLAGS, "level"},
        { "5",    "5 level(50)",   0, AV_OPT_TYPE_CONST, {.i64 = 50}, 0, 0, FLAGS, "level"},
        { "5.1",  "5.1 level(51)", 0, AV_OPT_TYPE_CONST, {.i64 = 51}, 0, 0, FLAGS, "level"},
        { "5.2",  "5.2 level(52)", 0, AV_OPT_TYPE_CONST, {.i64 = 52}, 0, 0, FLAGS, "level"},
        { "6",    "6 level(60)",   0, AV_OPT_TYPE_CONST, {.i64 = 60}, 0, 0, FLAGS, "level"},
        { "6.1",  "6.1 level(61)", 0, AV_OPT_TYPE_CONST, {.i64 = 61}, 0, 0, FLAGS, "level"},
        { "6.2",  "6.2 level(62)", 0, AV_OPT_TYPE_CONST, {.i64 = 62}, 0, 0, FLAGS, "level"},

    { "profile", "Set the encoding profile", OFFSETENCPARAM(enc_profile), AV_OPT_TYPE_INT, {.i64 = PROFILE_HEVC_MAIN}, -1, PROFILE_HEVC_MAIN10_INTRA, FLAGS, "profile"},
        { "auto",         "", 0,  AV_OPT_TYPE_CONST, {.i64 = PROFILE_AUTO},              0, 0, FLAGS, "profile"},
        { "main",         "", 0,  AV_OPT_TYPE_CONST, {.i64 = PROFILE_HEVC_MAIN},         0, 0, FLAGS, "profile"},
        { "main-intra",   "", 0,  AV_OPT_TYPE_CONST, {.i64 = PROFILE_HEVC_MAIN_INTRA},   0, 0, FLAGS, "profile"},
        { "main-10",       "", 0,  AV_OPT_TYPE_CONST, {.i64 = PROFILE_HEVC_MAIN10},       0, 0, FLAGS, "profile"},
        { "main-10-intra", "", 0,  AV_OPT_TYPE_CONST, {.i64 = PROFILE_HEVC_MAIN10_INTRA}, 0, 0, FLAGS, "profile"},

    { "control_rate", "Rate control mode", OFFSETENCPARAM(rate_control), AV_OPT_TYPE_INT, {.i64 = MODE_RC_AUTO}, -1, 4, FLAGS, "control_rate"},
        { "auto",     "auto(-1)",                         0,  AV_OPT_TYPE_CONST, {.i64 = MODE_RC_AUTO},     0, 0, FLAGS, "control_rate"},
        { "const_qp", "contant qp(0)",                    0,  AV_OPT_TYPE_CONST, {.i64 = MODE_RC_CONST_QP}, 0, 0, FLAGS, "control_rate"},
        { "cbr",      "constabt bit-rate(1)",             0,  AV_OPT_TYPE_CONST, {.i64 = MODE_RC_CBR},      0, 0, FLAGS, "control_rate"},
        { "vbr",      "variable bit-rate(2)",             0,  AV_OPT_TYPE_CONST, {.i64 = MODE_RC_VBR},      0, 0, FLAGS, "control_rate"},
        { "cvbr",     "constrained variable bit-rate(3)", 0,  AV_OPT_TYPE_CONST, {.i64 = MODE_RC_CVBR},     0, 0, FLAGS, "control_rate"},
        { "cabr",     "content adaptive bit-rate(4)",     0,  AV_OPT_TYPE_CONST, {.i64 = MODE_RC_CABR},     0, 0, FLAGS, "control_rate"},

    { "preset", "Encoder preset Possible values: (slow, medium, fast)", OFFSETENCPARAM(preset), AV_OPT_TYPE_STRING, {.str = "medium"}, .flags=FLAGS, .unit="preset"},

    { "bf", "Number of B-frames", OFFSETENCPARAM(num_b_frames), AV_OPT_TYPE_INT, { .i64 = -1}, -1, 3, FLAGS, "b_frames"},
    { "qp", "Quantization parameter", OFFSETENCPARAM(qp), AV_OPT_TYPE_INT, {.i64 = -1}, -1, 51, FLAGS, "qp"},
    { "min_bitrate", "Minimum bit rate", OFFSETENCPARAM(min_bitrate), AV_OPT_TYPE_INT64, {.i64 = -1}, -1, 35000000000, FLAGS, "min_bitrate"},
    { "max_bitrate", "Maximum bit rate", OFFSETENCPARAM(max_bitrate), AV_OPT_TYPE_INT64, {.i64 = -1}, -1, 35000000000, FLAGS, "max_bitrate"},
    { "slice", "Slice num to use for encoding", OFFSETENCPARAM(slice_id), AV_OPT_TYPE_INT, { .i64 = -1}, -1, 1, FLAGS, "slice"},
    { "cores", "Cores decide if it's 1 or 2-slice encoding", OFFSETENCPARAM(num_cores), AV_OPT_TYPE_INT, { .i64 = 1}, 1, 1, FLAGS, "cores"},

    { "crf", "Enable/Disable CRF mode", OFFSETENCPARAM(crf),  AV_OPT_TYPE_INT, {.i64 = 0},  0, 1, FLAGS, "crf"},
        { "disable", "", 0,  AV_OPT_TYPE_CONST, {.i64 = 0}, 0, 0, FLAGS, "crf"},
        { "enable", "",  0,  AV_OPT_TYPE_CONST, {.i64 = 1}, 0, 0, FLAGS, "crf"},

    { "forced_idr", "If forcing keyframes, force them as IDR frames.", OFFSETENCPARAM(enc_force_idr),  AV_OPT_TYPE_INT, {.i64 = 1}, 0, 1, FLAGS, "forced_idr"},
        { "disable", "", 0,  AV_OPT_TYPE_CONST, {.i64 = 0}, 0, 0, FLAGS, "forced_idr"},
        { "enable",  "", 0,  AV_OPT_TYPE_CONST, {.i64 = 1}, 0, 0, FLAGS, "forced_idr"},

    { "lookahead_depth", "Number of frames in pipeline prior to the encoder", OFFSETENCPARAM(la_depth), AV_OPT_TYPE_INT, {.i64 = -1},  -1, 53, FLAGS, "lookahead_depth"},
    { "latency_ms", "Lookahead depth specified in milliseconds", OFFSETENCPARAM(latency_ms), AV_OPT_TYPE_INT, {.i64 = -1},  -1, 60000, FLAGS, "latency_ms"},
    { "no_bll", "No bframes for low latency usecase", OFFSETENCPARAM(no_low_latency_b_frames), AV_OPT_TYPE_INT, {.i64 = -1},  -1, 1, FLAGS, "no_bll"},

    { "qp_mode", "Quantization parameter", OFFSETENCPARAM(qp_mode), AV_OPT_TYPE_INT, { .i64 = MODE_QP_AUTO}, 0, 2, FLAGS, "qp_mode"},
        { "auto",          "best VQ",                  0, AV_OPT_TYPE_CONST, {.i64 = MODE_QP_AUTO},     0, 0, FLAGS, "qp_mode"},
        { "relative_load", "spatial AQ + temporal AQ", 0, AV_OPT_TYPE_CONST, {.i64 = MODE_QP_REL_LOAD}, 0, 0, FLAGS, "qp_mode"},
        { "uniform",        "AQ off",                  0, AV_OPT_TYPE_CONST, {.i64 = MODE_QP_UNIFORM},  0, 0, FLAGS, "qp_mode"},

    { "bufsize", "Size of VBV buffer (in bits)", OFFSETENCPARAM(bufsize), AV_OPT_TYPE_INT, { .i64 = -1}, -1, 400000000, FLAGS, "bufsize"},

    { "spatial_aq", "Enable spatial AQ", OFFSETENCPARAM(spatial_aq), AV_OPT_TYPE_INT, {.i64 = -1}, -1, 1, FLAGS, "spatial_aq"},
        { "auto",    "", 0,  AV_OPT_TYPE_CONST, {.i64 = -1}, 0, 0, FLAGS, "spatial_aq"},
        { "disable", "", 0,  AV_OPT_TYPE_CONST, {.i64 = 0},  0, 0, FLAGS, "spatial_aq"},
        { "enable", "",  0,  AV_OPT_TYPE_CONST, {.i64 = 1},  0, 0, FLAGS, "spatial_aq"},

    { "temporal_aq", "Enable temporal AQ", OFFSETENCPARAM(temporal_aq), AV_OPT_TYPE_INT, {.i64 = -1}, -1, 1, FLAGS, "temporal_aq"},
        { "auto",    "", 0,  AV_OPT_TYPE_CONST, {.i64 = -1}, 0, 0, FLAGS, "temporal_aq"},
        { "disable", "", 0,  AV_OPT_TYPE_CONST, {.i64 = 0},  0, 0, FLAGS, "temporal_aq"},
        { "enable",  "", 0,  AV_OPT_TYPE_CONST, {.i64 = 1},  0, 0, FLAGS, "temporal_aq"},

    { "spatial_aq_gain",  "Spatial aq gain scale factor",  OFFSETENCPARAM(spatial_aq_gain),  AV_OPT_TYPE_INT, {.i64 = 255}, 0, 255, FLAGS, "spatial_aq_gain"},
        { "auto", "", 0,  AV_OPT_TYPE_CONST, {.i64 = 255}, 0, 0, FLAGS, "spatial_aq_gain"},
    { "temporal_aq_gain", "Temporal aq gain scale factor", OFFSETENCPARAM(temporal_aq_gain), AV_OPT_TYPE_INT, {.i64 = 255}, 0, 255, FLAGS, "temporal_aq_gain"},
        { "auto", "", 0,  AV_OPT_TYPE_CONST, {.i64 = 255}, 0, 0, FLAGS, "temporal_aq_gain"},

    { "min_qp", "Min Qp for rate control", OFFSETENCPARAM(min_qp), AV_OPT_TYPE_INT, {.i64 = -1}, -1, 51, FLAGS, "min_qp"},
    { "max_qp", "Min Qp for rate control", OFFSETENCPARAM(max_qp), AV_OPT_TYPE_INT, {.i64 = -1}, -1, 51, FLAGS, "max_qp"},
    { "qp_i_offset", "QP Offset - I", OFFSETENCPARAM(qp_i_offset), AV_OPT_TYPE_INT, {.i64 = 0}, -51, 51, FLAGS, "qp_i_offset"},
    { "qp_b_offset", "QP Offset - B/P", OFFSETENCPARAM(qp_b_offset), AV_OPT_TYPE_INT, {.i64 = 0}, -51, 51, FLAGS, "qp_b_offset"},

    { "tune_metrics", "Tune VQ for objective metrics", OFFSETENCPARAM(tune_metrics), AV_OPT_TYPE_INT, {.i64 = TM_BEST_VQ}, 1, 4, FLAGS, "tune_metrics"},
        { "vq",      "best VQ (1)",   0, AV_OPT_TYPE_CONST, {.i64 = TM_BEST_VQ},   0, 0, FLAGS, "tune_metrics"},
        { "psnr",    "best PSNR (2)", 0, AV_OPT_TYPE_CONST, {.i64 = TM_BEST_PSNR}, 0, 0, FLAGS, "tune_metrics"},
        { "ssim",    "best SSIM (3)", 0, AV_OPT_TYPE_CONST, {.i64 = TM_BEST_SSIM}, 0, 0, FLAGS, "tune_metrics"},
        { "vmaf",    "best VMAF (4)", 0, AV_OPT_TYPE_CONST, {.i64 = TM_BEST_VMAF}, 0, 0, FLAGS, "tune_metrics"},

    { "tier", "HEVC tier", OFFSETENCPARAM(enc_tier), AV_OPT_TYPE_INT, {.i64 = -1},  -1,  1, FLAGS, "tier"},
        { "auto", "", 0, AV_OPT_TYPE_CONST, {.i64 = TEIR_ENC_AUTO}, 0, 1, FLAGS, "tier"},
        { "main", "", 0, AV_OPT_TYPE_CONST, {.i64 = TEIR_ENC_MAIN}, 0, 1, FLAGS, "tier"},
        { "high", "", 0, AV_OPT_TYPE_CONST, {.i64 = TEIR_ENC_HIGH}, 0, 1, FLAGS, "tier"},

    { "latency_logging", "Log latency info to syslog", OFFSETENCPARAM(latency_logging), AV_OPT_TYPE_INT, {.i64 = 0}, 0, 1, FLAGS, "latency_logging"},
        { "disable", "", 0,  AV_OPT_TYPE_CONST, {.i64 = 0}, 0, 0, FLAGS, "latency_logging"},
        { "enable", "",  0,  AV_OPT_TYPE_CONST, {.i64 = 1}, 0, 0, FLAGS, "latency_logging"},

    { "dynamic_gop", "Dynamically determine the optimal number of B frames, based on the content.", OFFSETENCPARAM(dynamic_gop),  AV_OPT_TYPE_INT, {.i64 = -1}, -1, 1, FLAGS, "dynamic_gop"},
        { "auto",    "", 0,  AV_OPT_TYPE_CONST, {.i64 = -1}, 0, 0, FLAGS, "dynamic_gop"},
        { "disable", "", 0,  AV_OPT_TYPE_CONST, {.i64 = 0},  0, 0, FLAGS, "dynamic_gop"},
        { "enable",  "", 0,  AV_OPT_TYPE_CONST, {.i64 = 1},  0, 0, FLAGS, "dynamic_gop"},

    { "dynamic_params_file", "Configuration file to dynamically change codec params", OFFSET(dynamic_param_location), AV_OPT_TYPE_STRING, {.str = NULL}, .flags=FLAGS, .unit="dynamic_params_file"},

    { "expert_options", "encoder configuration using a ',' separated list of key=value parameters", OFFSETENCPARAM(xav1_options), AV_OPT_TYPE_STRING, {.str = NULL}, .flags=FLAGS, .unit="expert_options"},
    { NULL },
};


static const AVOption av1enc_options[] = {
    { "level", "Set the encoding level restriction", OFFSETENCPARAM(enc_level), AV_OPT_TYPE_INT, { .i64 = 0 }, 0, 63, FLAGS, "level" },
        { "auto", "0 Auto",        0, AV_OPT_TYPE_CONST, {.i64 = 0},  0, 0, FLAGS, "level"},
        { "2",    "2 level(20)",   0, AV_OPT_TYPE_CONST, {.i64 = 20}, 0, 0, FLAGS, "level"},
        { "2.1",  "2.1 level(21)", 0, AV_OPT_TYPE_CONST, {.i64 = 21}, 0, 0, FLAGS, "level"},
        { "3",    "3 level(30)",   0, AV_OPT_TYPE_CONST, {.i64 = 30}, 0, 0, FLAGS, "level"},
        { "3.1",  "3.1 level(31)", 0, AV_OPT_TYPE_CONST, {.i64 = 31}, 0, 0, FLAGS, "level"},
        { "4",    "4 level(40)",   0, AV_OPT_TYPE_CONST, {.i64 = 40}, 0, 0, FLAGS, "level"},
        { "4.1",  "4.1 level(41)", 0, AV_OPT_TYPE_CONST, {.i64 = 41}, 0, 0, FLAGS, "level"},
        { "5",    "5 level(50)",   0, AV_OPT_TYPE_CONST, {.i64 = 50}, 0, 0, FLAGS, "level"},
        { "5.1",  "5.1 level(51)", 0, AV_OPT_TYPE_CONST, {.i64 = 51}, 0, 0, FLAGS, "level"},
        { "5.2",  "5.2 level(52)", 0, AV_OPT_TYPE_CONST, {.i64 = 52}, 0, 0, FLAGS, "level"},
        { "5.3",  "5.3 level(53)", 0, AV_OPT_TYPE_CONST, {.i64 = 53}, 0, 0, FLAGS, "level"},
        { "6",    "6 level(60)",   0, AV_OPT_TYPE_CONST, {.i64 = 60}, 0, 0, FLAGS, "level"},
        { "6.1",  "6.1 level(61)", 0, AV_OPT_TYPE_CONST, {.i64 = 61}, 0, 0, FLAGS, "level"},
        { "6.2",  "6.2 level(62)", 0, AV_OPT_TYPE_CONST, {.i64 = 62}, 0, 0, FLAGS, "level"},
        { "6.3",  "6.3 level(63)", 0, AV_OPT_TYPE_CONST, {.i64 = 63}, 0, 0, FLAGS, "level"},


    { "profile", "Set the encoding profile", OFFSETENCPARAM(enc_profile), AV_OPT_TYPE_INT, {.i64 = PROFILE_AUTO}, -1, PROFILE_AV1_MAIN, FLAGS, "profile"},
        { "auto", "", 0,  AV_OPT_TYPE_CONST, {.i64 = PROFILE_AUTO},     0, 0, FLAGS, "profile"},
        { "main", "", 0,  AV_OPT_TYPE_CONST, {.i64 = PROFILE_AV1_MAIN}, 0, 0, FLAGS, "profile"},

    { "control_rate", "Rate control mode", OFFSETENCPARAM(rate_control), AV_OPT_TYPE_INT, {.i64 = MODE_RC_AUTO}, -1, 4, FLAGS, "control_rate"},
        { "auto",     "auto(-1)",                         0,  AV_OPT_TYPE_CONST, {.i64 = MODE_RC_AUTO},     0, 0, FLAGS, "control_rate"},
        { "const_qp", "contant qp(0)",                    0,  AV_OPT_TYPE_CONST, {.i64 = MODE_RC_CONST_QP}, 0, 0, FLAGS, "control_rate"},
        { "cbr",      "constabt bit-rate(1)",             0,  AV_OPT_TYPE_CONST, {.i64 = MODE_RC_CBR},      0, 0, FLAGS, "control_rate"},
        { "vbr",      "variable bit-rate(2)",             0,  AV_OPT_TYPE_CONST, {.i64 = MODE_RC_VBR},      0, 0, FLAGS, "control_rate"},
        { "cvbr",     "constrained variable bit-rate(3)", 0,  AV_OPT_TYPE_CONST, {.i64 = MODE_RC_CVBR},     0, 0, FLAGS, "control_rate"},
        { "cabr",     "content adaptive bit-rate(4)",     0,  AV_OPT_TYPE_CONST, {.i64 = MODE_RC_CABR},     0, 0, FLAGS, "control_rate"},

    { "preset", "Encoder preset Possible values: (slow, medium, fast)", OFFSETENCPARAM(preset), AV_OPT_TYPE_STRING, {.str = "medium"}, .flags=FLAGS, .unit="preset"},

    { "bf", "Number of B-frames", OFFSETENCPARAM(num_b_frames), AV_OPT_TYPE_INT, { .i64 = -1}, -1, 7, FLAGS, "b_frames"},
    { "qp", "Quantization parameter", OFFSETENCPARAM(qp), AV_OPT_TYPE_INT, {.i64 = -1}, -1, 255, FLAGS, "qp"},
    { "min_bitrate", "Minimum bit rate", OFFSETENCPARAM(min_bitrate), AV_OPT_TYPE_INT64, {.i64 = -1}, -1, 35000000000, FLAGS, "min_bitrate"},
    { "max_bitrate", "Maximum bit rate", OFFSETENCPARAM(max_bitrate), AV_OPT_TYPE_INT64, {.i64 = -1}, -1, 35000000000, FLAGS, "max_bitrate"},
    { "slice", "Slice num to use for encoding", OFFSETENCPARAM(slice_id), AV_OPT_TYPE_INT, { .i64 = -1}, -1, 1, FLAGS, "slice"},
    { "cores", "Cores decide if it's 1 or 2-slice encoding", OFFSETENCPARAM(num_cores), AV_OPT_TYPE_INT, { .i64 = 1}, 1, 2, FLAGS, "cores"},

    { "crf", "Enable/Disable CRF mode", OFFSETENCPARAM(crf),  AV_OPT_TYPE_INT, {.i64 = 0},  0, 1, FLAGS, "crf"},
        { "disable", "", 0,  AV_OPT_TYPE_CONST, {.i64 = 0}, 0, 0, FLAGS, "crf"},
        { "enable", "",  0,  AV_OPT_TYPE_CONST, {.i64 = 1}, 0, 0, FLAGS, "crf"},

    { "forced_idr", "If forcing keyframes, force them as IDR frames.", OFFSETENCPARAM(enc_force_idr), AV_OPT_TYPE_INT, {.i64 = 1}, 0, 1, FLAGS, "forced_idr"},
        { "disable", "", 0,  AV_OPT_TYPE_CONST,{.i64 = 0}, 0, 0, FLAGS, "forced_idr"},
        { "enable",  "", 0,  AV_OPT_TYPE_CONST,{.i64 = 1}, 0, 0, FLAGS, "forced_idr"},

    { "lookahead_depth", "Number of frames in pipeline prior to the encoder", OFFSETENCPARAM(la_depth), AV_OPT_TYPE_INT, {.i64 = -1},  -1, 53, FLAGS, "lookahead_depth"},
    { "latency_ms", "Lookahead depth specified in milliseconds", OFFSETENCPARAM(latency_ms), AV_OPT_TYPE_INT, {.i64 = -1},  -1, 60000, FLAGS, "latency_ms"},
    { "no_bll", "No bframes for low latency usecase", OFFSETENCPARAM(no_low_latency_b_frames), AV_OPT_TYPE_INT, {.i64 = -1},  -1, 1, FLAGS, "no_bll"},

    { "qp_mode", "Quantization parameter", OFFSETENCPARAM(qp_mode), AV_OPT_TYPE_INT, { .i64 = MODE_QP_AUTO}, 0, 2, FLAGS, "qp_mode"},
        { "auto",          "best VQ",                  0, AV_OPT_TYPE_CONST, {.i64 = MODE_QP_AUTO},     0, 0, FLAGS, "qp_mode"},
        { "relative_load", "spatial AQ + temporal AQ", 0, AV_OPT_TYPE_CONST, {.i64 = MODE_QP_REL_LOAD}, 0, 0, FLAGS, "qp_mode"},
        { "uniform",        "AQ off",                  0, AV_OPT_TYPE_CONST, {.i64 = MODE_QP_UNIFORM},  0, 0, FLAGS, "qp_mode"},

    { "bufsize", "Size of VBV buffer (in bits)", OFFSETENCPARAM(bufsize), AV_OPT_TYPE_INT, { .i64 = -1}, -1, 400000000, FLAGS, "bufsize"},

    { "spatial_aq", "Enable spatial AQ", OFFSETENCPARAM(spatial_aq), AV_OPT_TYPE_INT, {.i64 = -1}, -1, 1, FLAGS, "spatial_aq"},
        { "auto",    "", 0,  AV_OPT_TYPE_CONST, {.i64 = -1}, 0, 0, FLAGS, "spatial_aq"},
        { "disable", "", 0,  AV_OPT_TYPE_CONST, {.i64 = 0},  0, 0, FLAGS, "spatial_aq"},
        { "enable", "",  0,  AV_OPT_TYPE_CONST, {.i64 = 1},  0, 0, FLAGS, "spatial_aq"},

    { "temporal_aq", "Enable temporal AQ", OFFSETENCPARAM(temporal_aq), AV_OPT_TYPE_INT, {.i64 = -1}, -1, 1, FLAGS, "temporal_aq"},
        { "auto",    "", 0,  AV_OPT_TYPE_CONST, {.i64 = -1}, 0, 0, FLAGS, "temporal_aq"},
        { "disable", "", 0,  AV_OPT_TYPE_CONST, {.i64 = 0},  0, 0, FLAGS, "temporal_aq"},
        { "enable",  "", 0,  AV_OPT_TYPE_CONST, {.i64 = 1},  0, 0, FLAGS, "temporal_aq"},

    { "spatial_aq_gain",  "Spatial aq gain scale factor",  OFFSETENCPARAM(spatial_aq_gain),  AV_OPT_TYPE_INT, {.i64 = 255}, 0, 255, FLAGS, "spatial_aq_gain"},
        { "auto", "", 0,  AV_OPT_TYPE_CONST, {.i64 = 255}, 0, 0, FLAGS, "spatial_aq_gain"},
    { "temporal_aq_gain", "Temporal aq gain scale factor", OFFSETENCPARAM(temporal_aq_gain), AV_OPT_TYPE_INT, {.i64 = 255}, 0, 255, FLAGS, "temporal_aq_gain"},
        { "auto", "", 0,  AV_OPT_TYPE_CONST, {.i64 = 255}, 0, 0, FLAGS, "temporal_aq_gain"},

    { "min_qp", "Min Qp for rate control", OFFSETENCPARAM(min_qp), AV_OPT_TYPE_INT, {.i64 = -1}, -1, 255, FLAGS, "min_qp"},
    { "max_qp", "Min Qp for rate control", OFFSETENCPARAM(max_qp), AV_OPT_TYPE_INT, {.i64 = -1}, -1, 255, FLAGS, "max_qp"},
    { "qp_i_offset", "QP Offset - I", OFFSETENCPARAM(qp_i_offset), AV_OPT_TYPE_INT, {.i64 = 0}, -51, 51, FLAGS, "qp_i_offset"},
    { "qp_b_offset", "QP Offset - B/P", OFFSETENCPARAM(qp_b_offset), AV_OPT_TYPE_INT, {.i64 = 0}, -51, 51, FLAGS, "qp_b_offset"},

    { "tune_metrics", "Tune VQ for objective metrics", OFFSETENCPARAM(tune_metrics), AV_OPT_TYPE_INT, {.i64 = TM_BEST_VQ}, 1, 4, FLAGS, "tune_metrics"},
        { "vq",      "best VQ (1)",   0, AV_OPT_TYPE_CONST, {.i64 = TM_BEST_VQ},   0, 0, FLAGS, "tune_metrics"},
        { "psnr",    "best PSNR (2)", 0, AV_OPT_TYPE_CONST, {.i64 = TM_BEST_PSNR}, 0, 0, FLAGS, "tune_metrics"},
        { "ssim",    "best SSIM (3)", 0, AV_OPT_TYPE_CONST, {.i64 = TM_BEST_SSIM}, 0, 0, FLAGS, "tune_metrics"},
        { "vmaf",    "best VMAF (4)", 0, AV_OPT_TYPE_CONST, {.i64 = TM_BEST_VMAF}, 0, 0, FLAGS, "tune_metrics"},

    { "tier", "AV1 tier", OFFSETENCPARAM(enc_tier), AV_OPT_TYPE_INT, {.i64 = -1},  -1,  1, FLAGS, "tier"},
        { "auto", "", 0, AV_OPT_TYPE_CONST, {.i64 = TEIR_ENC_AUTO}, 0, 1, FLAGS, "tier"},
        { "main", "", 0, AV_OPT_TYPE_CONST, {.i64 = TEIR_ENC_MAIN}, 0, 1, FLAGS, "tier"},
        { "high", "", 0, AV_OPT_TYPE_CONST, {.i64 = TEIR_ENC_HIGH}, 0, 1, FLAGS, "tier"},

    { "latency_logging", "Log latency info to syslog", OFFSETENCPARAM(latency_logging), AV_OPT_TYPE_INT, {.i64 = 0}, 0, 1, FLAGS, "latency_logging"},
        { "disable", "", 0,  AV_OPT_TYPE_CONST, {.i64 = 0}, 0, 0, FLAGS, "latency_logging"},
        { "enable",  "", 0,  AV_OPT_TYPE_CONST, {.i64 = 1}, 0, 0, FLAGS, "latency_logging"},

    { "type", "Select encoder type", OFFSETENCPARAM(av1_type), AV_OPT_TYPE_INT, {.i64 = 1},  1,  2, FLAGS, "type"},

    { "dynamic_gop", "Dynamically determine the optimal number of B frames, based on the content.", OFFSETENCPARAM(dynamic_gop),  AV_OPT_TYPE_INT, {.i64 = -1},  -1, 1, FLAGS, "dynamic_gop"},
        { "auto",    "", 0,  AV_OPT_TYPE_CONST, {.i64 = -1}, 0, 0, FLAGS, "dynamic_gop"},
        { "disable", "", 0,  AV_OPT_TYPE_CONST, {.i64 = 0},  0, 0, FLAGS, "dynamic_gop"},
        { "enable", "",  0,  AV_OPT_TYPE_CONST, {.i64 = 1},  0, 0, FLAGS, "dynamic_gop"},

    { "dynamic_params_file", "Configuration file to dynamically change codec params", OFFSET(dynamic_param_location), AV_OPT_TYPE_STRING, {.str = NULL}, .flags=FLAGS, .unit="dynamic_params_file"},

    { "expert_options", "encoder configuration using a ',' separated list of key=value parameters", OFFSETENCPARAM(xav1_options), AV_OPT_TYPE_STRING, {.str = NULL}, .flags=FLAGS, .unit="expert_options"},
    { NULL },
};

static int named_intconst_to_string(const snEncContext *ctx, const char *unit, int val, char** str_val)
{
    int ret = 0;
    const AVOption* opt = av_opt_next(ctx, NULL);
    while(opt){
        if(opt->type == AV_OPT_TYPE_CONST && !strcmp(opt->unit, unit) && opt->default_val.i64==val){
            if(*str_val)
                free(*str_val);
            *str_val = (char*)malloc(strlen(opt->name)+1);
            if(!str_val){
               av_log(NULL, AV_LOG_ERROR, "[%s] unale to allocated memory\n", __FUNCTION__);
               ret = -1;
               break;
            }
            strcpy(*str_val, opt->name);
            ret = 0;
            break;
        }
        opt = av_opt_next(ctx, opt);
    }
    return ret;
}

static void snenc_queue_pts (AVFifo* queue, int64_t pts)
{
    if (av_fifo_can_write(queue))
        av_fifo_write(queue, &pts, 1);
    else
        av_log(NULL, AV_LOG_ERROR, "[%s] Unable to queue pts\n", __FUNCTION__);

}

static int64_t snenc_dequeue_pts(AVFifo* queue)
{
    int64_t ts = AV_NOPTS_VALUE;
    if (av_fifo_can_read(queue)){
        av_fifo_read(queue, &ts, 1);
    }
     else
        av_log(NULL, AV_LOG_ERROR, "[%s] Unable to dequeue pts\n", __FUNCTION__);
    return ts;
}

static void snenc_prepare_out_timestamp (AVCodecContext *avctx, AVPacket *pkt)
{
    snEncContext *ctx = (snEncContext *)avctx->priv_data;

    // TODO: code is written based on nvenc, need to check legalities of this
    pkt->dts = snenc_dequeue_pts(ctx->pts_queue);
}

static void set_instance_name(AVCodecContext *avctx)
{
    snEncContext *enc_ctx = avctx->priv_data;
    snprintf(enc_ctx->instance_name, 64, "%s@%p", (*(AVClass **)avctx)->item_name(avctx), avctx);
}

static int snenc_init_hwctx(AVCodecContext *avctx)
{
    snEncContext *ctx = (snEncContext *)avctx->priv_data;
    AVVpeDeviceContext *device_hwctx;
    AVVpeFramesContext *frame_hwctx;
    AVHWFramesContext  *hwframe_ctx;
    AVHWDeviceContext  *hwdevice_ctx;
    AVVpeHWConfig      *hwconfig = NULL;
    AVHWFramesConstraints *constraints = NULL;
    int ret = 0;
    int i;

    /*Get HW frame. The avctx->hw_frames_ctx is the reference to the
      AVHWFramesContext describing the input frame for vpe encoder*/
   if (avctx->hw_frames_ctx) {
        ctx->hwframe = av_buffer_ref(avctx->hw_frames_ctx);
        if (!ctx->hwframe) {
            av_log(avctx, AV_LOG_DEBUG, "hwframe ref failed\n");
            return AVERROR(ENOMEM);
        }
    } else {
        av_log(avctx, AV_LOG_DEBUG, "hw_frames_ctx NULL\n");
        return AVERROR_INVALIDDATA;
    }

    hwframe_ctx = (AVHWFramesContext*)ctx->hwframe->data;
    if (avctx->hw_device_ctx) {
        hwconfig = av_hwdevice_hwconfig_alloc(avctx->hw_device_ctx);
    } else {
        ctx->hwdevice = av_buffer_ref(hwframe_ctx->device_ref);
        if (!ctx->hwdevice) {
            av_log(avctx, AV_LOG_DEBUG, "hwdevice ref failed\n");
            return AVERROR(ENOMEM);
        }
        hwconfig = av_hwdevice_hwconfig_alloc(ctx->hwdevice);
    }
    if (!hwconfig) {
        av_log(avctx, AV_LOG_DEBUG, "hwconfig NULL\n");
        return AVERROR(ENOMEM);
    }

    switch (avctx->codec_id) {
        case AV_CODEC_ID_H264:
            hwconfig->id  = VPE_HW_H264ENC;
            break;
        case AV_CODEC_ID_HEVC:
            hwconfig->id  = VPE_HW_HEVCENC;
            break;
        case AV_CODEC_ID_AV1:
            hwconfig->id  = VPE_HW_XAV1ENC;
            break;
        default:
            ret = AVERROR(EINVAL);
            goto error_exit;
    }
    av_log(avctx, AV_LOG_DEBUG, "hwconfig->id = %d\n", hwconfig->id);

    if (avctx->hw_device_ctx) {
        constraints =
            av_hwdevice_get_hwframe_constraints(avctx->hw_device_ctx, hwconfig);
    } else {
        constraints =
            av_hwdevice_get_hwframe_constraints(ctx->hwdevice, hwconfig);
    }
    if (!constraints) {
        ret = AVERROR(ENOMEM);
        goto error_exit;
    }

    av_log(avctx, AV_LOG_DEBUG, "get constraints %d,%d,%d,%d\n", constraints->min_width,
                    constraints->min_height, constraints->max_width, constraints->max_height);

    if (avctx->coded_width  < constraints->min_width  ||
        avctx->coded_height < constraints->min_height ||
        avctx->coded_width  > constraints->max_width  ||
        avctx->coded_height > constraints->max_height) {
        av_log(avctx, AV_LOG_ERROR, "Hardware does not support image "
               "size %dx%d (constraints: width %d-%d height %d-%d).\n",
               avctx->coded_width, avctx->coded_height,
               constraints->min_width,  constraints->max_width,
               constraints->min_height, constraints->max_height);
        ret = AVERROR(EINVAL);
        goto error_exit;
    }

    if (!constraints->valid_sw_formats ||
        constraints->valid_sw_formats[0] == AV_PIX_FMT_NONE) {
        av_log(avctx, AV_LOG_ERROR, "Hardware does not offer any "
               "usable surface formats.\n");
        ret = AVERROR(EINVAL);
        goto error_exit;
    }

    av_log(avctx, AV_LOG_DEBUG, "avctx->sw_pix_fmt %s\n",
                                 av_get_pix_fmt_name(avctx->sw_pix_fmt));
    av_log(avctx, AV_LOG_DEBUG, "hwframe_ctx->sw_format %s\n",
                                 av_get_pix_fmt_name(hwframe_ctx->sw_format));
    i = 0;
    while (1) {
        if (avctx->sw_pix_fmt == constraints->valid_sw_formats[i])
            break;
        if (constraints->valid_sw_formats[i] == AV_PIX_FMT_NONE) {
            av_log(ctx, AV_LOG_ERROR, "Hardware can't find any matched sw format\n");
            ret = AVERROR(EINVAL);
            goto error_exit;
        }
        i++;
    }
    av_log(avctx, AV_LOG_DEBUG, "get sw constraints %s\n",
                                 av_get_pix_fmt_name(constraints->valid_sw_formats[i]));

    if (!constraints->valid_hw_formats ||
        constraints->valid_hw_formats[0] != AV_PIX_FMT_VPE) {
        av_log(avctx, AV_LOG_ERROR, "Hardware error format\n");
        ret = AVERROR(EINVAL);
        goto error_exit;
    }
    av_log(avctx, AV_LOG_DEBUG, "get hw constraints %s\n",
                                 av_get_pix_fmt_name(constraints->valid_hw_formats[0]));

    if (!avctx->hw_device_ctx) {
        device_hwctx = hwframe_ctx->device_ctx->hwctx;
    } else {
        hwdevice_ctx = (AVHWDeviceContext *)avctx->hw_device_ctx->data;
        device_hwctx = (AVVpeDeviceContext *)hwdevice_ctx->hwctx;
    }
    frame_hwctx  = hwframe_ctx->hwctx;

    vpi_device_set_ctx(&ctx->enc_params, device_hwctx->vpe_dev_ctx, VPE_HW_ENC);
    ctx->enc_params.mem_ctx = frame_hwctx->vpe_mem_ctx;

    av_log(avctx, AV_LOG_DEBUG, "coded param: %d x %d\n", avctx->coded_width, avctx->coded_height);
    av_log(avctx, AV_LOG_DEBUG, "hwframe_ctx->format = %d\n", hwframe_ctx->format);
    av_log(avctx, AV_LOG_DEBUG, "hwframe_ctx->sw_format = %d\n", hwframe_ctx->sw_format);

    ctx->enc_params.orig_width  = avctx->width;
    ctx->enc_params.orig_height = avctx->height;
    ctx->enc_params.width       = avctx->width;
    ctx->enc_params.height      = avctx->height;

    if (hwframe_ctx->sw_format == AV_PIX_FMT_P010LE ||
        hwframe_ctx->sw_format == AV_PIX_FMT_YUV420P10LE ||
        hwframe_ctx->sw_format == AV_PIX_FMT_PACKED10) {
        ctx->enc_params.bitdepth = 10;
    } else {
        ctx->enc_params.bitdepth = 8;
    }

    av_hwframe_constraints_free(&constraints);
    av_freep(&hwconfig);

    return 0;

error_exit:
    if (constraints) {
        av_hwframe_constraints_free(&constraints);
    }
    if (hwconfig) {
        av_freep(&hwconfig);
    }

    return ret;
}

/**
 * Init vpe enc hw ctx and create vpe context
 */
static av_cold int _sn_encode_init(AVCodecContext *avctx)
{
    snEncContext *ctx = (snEncContext *)avctx->priv_data;
    VpiEncInitParams *enc_params = (VpiEncInitParams*)&ctx->enc_params;
    const AVPixFmtDescriptor *desc;
    bool is_xav1;
    int ret, i;
#if CONFIG_LIBXRM
    XrmInterfaceProperties xrm_props;
    int dev_index = -1;
    bool is_ull;
    memset(&xrm_props, 0, sizeof(xrm_props));
#endif
    ret = snenc_init_hwctx(avctx);
    if (ret) {
        av_log(avctx, AV_LOG_ERROR, "snenc init hwctx failed\n");
        return ret;
    }
    is_xav1 = strcmp(avctx->codec->name, "av1_ama") == 0 && enc_params->av1_type == VPI_ENC_TYPE_1;

    if(enc_params->num_cores == 2) {
        if(!is_xav1) {
            av_log(avctx, AV_LOG_ERROR, "2-slice encoding only supported for type 1 AV1 encoder. \n");
            return AVERROR(EINVAL);
        }
        if(enc_params->la_depth != 0) {
            av_log(avctx, AV_LOG_ERROR, "For 2-slice encoding, lookahead depth should be set to 0. \n");
            return AVERROR(EINVAL);
        }
        if(enc_params->slice_id != -1) {
            av_log(avctx, AV_LOG_ERROR, "For 2-slice encoding, slice id needs to be set internally. Please set it to -1 \n");
            return AVERROR(EINVAL);
        }
    }

#if CONFIG_LIBXRM
    // Reserve xrm resource
    xrm_props.width = avctx->width;
    xrm_props.height = avctx->height;
    xrm_props.fps_num = avctx->framerate.num;
    xrm_props.fps_den = avctx->framerate.den;
    xrm_props.is_la_enabled = enc_params->la_depth != 0; // TODO becomes non trival with latency_ms
    xrm_props.enc_cores = enc_params->num_cores;
    strncpy(xrm_props.preset, enc_params->preset, sizeof(xrm_props.preset)-1);

    if(vpi_device_get_device_index(ctx->enc_params.hw_ctx, &dev_index) < 0) {
        return AVERROR(EINVAL);
    }
    is_ull = enc_params->latency_mode == MODE_LATENCY_ULL;
    ret = xrm_enc_reserve(&ctx->xrm_ctx, dev_index, enc_params->slice_id, is_xav1, is_ull, &xrm_props);
    if (ret == XRM_ERROR) {
        return AVERROR(EBUSY);
    }
    if (ret == XRM_SUCCESS) {
        enc_params->slice_id = ctx->xrm_ctx.slice_id;
    }
#endif
    av_log(avctx, AV_LOG_VERBOSE, "slice id = %d\n", enc_params->slice_id);

    if (ctx->dynamic_param_location) {
        ctx->dynamic_params_fp = fopen(ctx->dynamic_param_location, "r");
        if (!ctx->dynamic_params_fp) {
            av_log(avctx, AV_LOG_ERROR, "Unable to open dynamic params file %s!\n", ctx->dynamic_param_location);
            return AVERROR(EINVAL);
        }
    } else {
        ctx->dynamic_params_fp = NULL;
    }
    ctx->sendFrameNum    = 0;
    ctx->recvFrameNum    = 0;
    ctx->eof_flush_en    = 0;
    ctx->flush_done      = 0;

    desc = av_pix_fmt_desc_get(avctx->sw_pix_fmt);
    if(!desc){
        av_log(avctx, AV_LOG_ERROR, "Unable to get input pixel format\n");
        return AVERROR_INVALIDDATA;

    }
    ctx->nb_components = desc->nb_components;
    if(ctx->nb_components < 1 || ctx->nb_components>3){
        av_log(avctx, AV_LOG_ERROR, "Invalid number of components in input format to encoder (%d) sw_fmt = %d\n", ctx->nb_components, avctx->sw_pix_fmt);
        return AVERROR(EINVAL);
    }
    for(i=0; i< ctx->nb_components; i++){
        ctx->sum_psnr[i] = 0;
        ctx->sum_ssim[i] = 0;
        ctx->max_psnr[i] = 0;
        ctx->max_ssim[i] = 0;
        ctx->min_psnr[i] = DBL_MAX;
        ctx->min_ssim[i] = DBL_MAX;
    }

    if(enc_params->av1_type == VPI_ENC_TYPE_2 && enc_params->num_b_frames > 3){
        av_log(avctx, AV_LOG_ERROR, "max_b_frames for encoder_type2 = 3, provided = %d\n", enc_params->num_b_frames);
        return AVERROR(EINVAL);
    }

    //set encoder params
    enc_params->enc_type       = VPE_ENCODER_XAV1;
    enc_params->rate_num       = avctx->framerate.num;
    enc_params->rate_denom     = avctx->framerate.den;
    enc_params->bit_rate       = avctx->bit_rate;
    enc_params->gop_size       = avctx->gop_size;

    ret = named_intconst_to_string(ctx, "level", ctx->enc_params.enc_level, &ctx->enc_params.level);
    if(ret < 0) {
        av_log(avctx, AV_LOG_ERROR, "named_intconst_to_string failed for level=%d\n", ctx->enc_params.enc_level);
        return AVERROR(EINVAL);
    }
    ret = named_intconst_to_string(ctx, "profile", ctx->enc_params.enc_profile, &ctx->enc_params.profile);
    if(ret < 0) {
        av_log(avctx, AV_LOG_ERROR, "named_intconst_to_string failed for profile=%d\n", ctx->enc_params.enc_profile);
        return AVERROR(EINVAL);
    }

    if (!strcmp(avctx->codec->name, "h264_ama")) {
        enc_params->codec_format = VPE_ENC_H264;
    } else if (!strcmp(avctx->codec->name, "hevc_ama")) {
        enc_params->codec_format = VPE_ENC_HEVC;
    } else if (!strcmp(avctx->codec->name, "av1_ama")) {
        enc_params->codec_format = enc_params->av1_type == VPI_ENC_TYPE_1 ? VPE_ENC_AV1 : VPE_ENC_AV1_T2;
    } else {
        av_log(avctx, AV_LOG_ERROR, "%s encoder not supported, please check your params!\n", avctx->codec->name);
        return AVERROR(EINVAL);
    }

    if(enc_params->no_low_latency_b_frames == 0) {
        av_log(avctx, AV_LOG_ERROR, "Invalid value for no_low_latency_b_frames=%d\n", ctx->enc_params.no_low_latency_b_frames);
        return AVERROR(EINVAL);
    }

    ctx->frame = av_frame_alloc();
    if (!ctx->frame)
        return AVERROR(ENOMEM);

    ctx->vpi_pkt = vpi_memory_get_struct(VPE_PACKET);
    if (!ctx->vpi_pkt)
        return AVERROR(ENOMEM);

    ctx->pts_queue = av_fifo_alloc2(128, sizeof(int64_t), 0);
    if (!ctx->pts_queue){
        av_log(avctx, AV_LOG_ERROR, "Unable to allocate memory for dts queue");
        return AVERROR(ENOMEM);
    }

    ret = sn_encode_start(avctx);
    if (ret) {
        av_log(ctx, AV_LOG_ERROR, "sn_encode_start failed\n");
        return ret;
    }

    return 0;
}

static av_cold int sn_encode_init(AVCodecContext *avctx)
{
    int ret;
    snEncContext *enc_ctx = avctx->priv_data;
    set_instance_name(avctx);
    {
        LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_ENC, enc_ctx->instance_name);
        ret = _sn_encode_init(avctx);
        LOG_AMA_RESTORE_INSTANCE_INFO();
    }
    return ret;
}

static int sn_encode_start(AVCodecContext *avctx)
{
    snEncContext *ctx = (snEncContext *)avctx->priv_data;
    AVHWFramesContext *hwframe_ctx;
    AVVpeFramesContext *frame_hwctx;
    int out_index, ret;
    int ch_format;
    int ch_index;
    VpiFrm *frame;
    int delay_pts;
    AVFrame *first_pic;
    int side_size;
    VpiSidedataHandle sb_handle;
    void *side_data;

    first_pic = (AVFrame *)avctx->opaque;
    if(!first_pic){
        av_log(avctx, AV_LOG_ERROR, "empty first_pic\n");
        return AVERROR_EXTERNAL;
    }

    hwframe_ctx = (AVHWFramesContext*)ctx->hwframe->data;
    frame_hwctx = hwframe_ctx->hwctx;
    frame = (VpiFrm *)first_pic->data[0];

    ret = vpi_memory_get_channel_info(frame_hwctx->vpe_mem_ctx, &frame_hwctx->proc_info->ch_info[0]);
    if (ret < 0) {
        av_log(avctx, AV_LOG_ERROR, "vpi_memory_get_info failed\n");
        return AVERROR_EXTERNAL;
    }

    if (first_pic->buf[1]) {
        out_index = *(int *)(first_pic->buf[1]->data);
    } else {
        out_index = 0;
    }

    ret = vpi_memory_get_map_info(frame_hwctx->vpe_mem_ctx, out_index, &ch_index);
    if (ret < 0) {
        av_log(avctx, AV_LOG_ERROR, "vpi_memory_get_map_info failed\n");
        return AVERROR_EXTERNAL;
    }

    ctx->enc_params.enc_index = out_index;
    ctx->enc_params.ch_index  = ch_index;

    av_log(avctx, AV_LOG_DEBUG, "frame_hwctx->proc_info->ch_info[%d].flag = %d\n", ch_index, frame_hwctx->proc_info->ch_info[ch_index].flag);
    av_log(avctx, AV_LOG_DEBUG, "frame_hwctx->proc_info->ch_info[%d].format = %d\n", ch_index, frame_hwctx->proc_info->ch_info[ch_index].format);
    av_log(avctx, AV_LOG_DEBUG, "frame_hwctx->proc_info->ch_info[%d].sw_format = %d\n", ch_index, frame_hwctx->proc_info->ch_info[ch_index].sw_format);

    //HDR related
    vpifrm_get_first_sidedata_by_type(frame, VPI_SB_VUI_PARAMS, &sb_handle);
    if(sb_handle) {
        side_size = vpi_get_sidedata_size(sb_handle);
        side_data = vpi_get_sidedata_vaddr(sb_handle);
        memcpy(&ctx->enc_params.vui_param, side_data, side_size);
    }

    if (ctx->enc_params.input_format == VPI_FORMAT_P010LE ||
        ctx->enc_params.input_format == VPI_FORMAT_YUV420P10LE ||
        ctx->enc_params.input_format == VPI_FORMAT_SP1010) {
        vpifrm_get_first_sidedata_by_type(frame, VPI_SB_SEI_MASTERING_DISPLAY_COLOR_VOLUME, &sb_handle);
        if(sb_handle) {
            side_size = vpi_get_sidedata_size(sb_handle);
            side_data = vpi_get_sidedata_vaddr(sb_handle);
            memcpy(&ctx->enc_params.mdcv, side_data, side_size);
        }

        vpifrm_get_first_sidedata_by_type(frame, VPI_SB_SEI_CONTENT_LIGHT_LEVEL, &sb_handle);
        if(sb_handle) {
            side_size = vpi_get_sidedata_size(sb_handle);
            side_data = vpi_get_sidedata_vaddr(sb_handle);
            memcpy(&ctx->enc_params.cll, side_data, side_size);
        }

        vpifrm_get_first_sidedata_by_type(frame, VPI_SB_SEI_ATC, &sb_handle);
        if(sb_handle) {
            side_size = vpi_get_sidedata_size(sb_handle);
            side_data = vpi_get_sidedata_vaddr(sb_handle);
            memcpy(&ctx->enc_params.atc, side_data, side_size);
        }
    }

    ctx->enc_params.walign    = frame_hwctx->proc_info->ch_info[ch_index].walign;
    ctx->enc_params.halign    = frame_hwctx->proc_info->ch_info[ch_index].halign;
    ctx->enc_params.flag      = frame_hwctx->proc_info->ch_info[ch_index].flag;

    ch_format = frame_hwctx->proc_info->ch_info[ch_index].format;
    if (ch_format == VPI_FORMAT_VPE) {
        ctx->enc_params.input_format = frame_hwctx->proc_info->ch_info[ch_index].sw_format;
    } else {
        av_log(avctx, AV_LOG_ERROR, "input format not support, please check your params!\n");
        return AVERROR_INVALIDDATA;
    }

    av_log(avctx, AV_LOG_DEBUG, "w x h = %d x %d, hwframe_ctx->sw_format = %d, format is %d, ch_format = %d\n",
                                 ctx->enc_params.width, ctx->enc_params.height, hwframe_ctx->sw_format, ctx->enc_params.input_format, ch_format);
    av_log(avctx, AV_LOG_DEBUG, "ch_w x ch_h = %d x %d\n",
                                frame_hwctx->proc_info->ch_info[ch_index].width, frame_hwctx->proc_info->ch_info[ch_index].height);


    av_log(avctx, AV_LOG_DEBUG, "input_format = %d\n", ctx->enc_params.input_format);

    ctx->enc_handle = vpi_encoder_init(&ctx->enc_params);
    if (ctx->enc_handle == NULL) {
        av_log(avctx, AV_LOG_ERROR, "vpi_encoder_init failed, error=%s\n",
                                     vpi_error_str(ret));
        return AVERROR_EXTERNAL;
    }

    ret = vpi_encoder_open(ctx->enc_handle);
    if (ret) {
        av_log(avctx, AV_LOG_ERROR, "vpi_encoder_open failed, error=%s\n",
                                     vpi_error_str(ret));
        return AVERROR_EXTERNAL;
    }

    ret = vpi_encoder_get_reorder_delay(ctx->enc_handle, &ctx->dts_delay);
    if (ret) {
        av_log(avctx, AV_LOG_ERROR, "vpi_encoder_get_reorder_delay failed, error=%s(%d)\n",
                                     vpi_error_str(ret), ret);
        return AVERROR_EXTERNAL;
    }
    av_log(avctx, AV_LOG_DEBUG, "reorder_delay = %u\n", ctx->dts_delay);

    //Handle non-contigious pts values while computing dts.
    delay_pts = -1*ctx->dts_delay;
    for(;delay_pts<0;delay_pts++){
        snenc_queue_pts(ctx->pts_queue, delay_pts* FFMAX(avctx->ticks_per_frame, 1));
    }


    if(!avctx->extradata_size) {
        char *extradata = (char *)vpi_encoder_get_extradata(ctx->enc_handle, &avctx->extradata_size);
        if (extradata) {
            avctx->extradata = av_mallocz(avctx->extradata_size + AV_INPUT_BUFFER_PADDING_SIZE);
            memcpy(avctx->extradata, extradata, avctx->extradata_size);
        }
    }


    avctx->opaque = NULL;

    ctx->initialized = 1;

    return 0;
}

/**
 * Fetch new AVFrame and put to external encoder
 */
static int sn_encode_send_frame(AVCodecContext *avctx)
{
    snEncContext *ctx = (snEncContext *)avctx->priv_data;
    int ret;
    VpiFrm *vpi_frame    = NULL;
    AVFrame *frame       = ctx->frame;
    AVFrame *trans_frame = NULL;

    if (!frame->buf[0]) {
        ret = ff_encode_get_frame(avctx, frame);
        if (ret < 0 && ret != AVERROR_EOF)
            return ret;
        if (ret == AVERROR_EOF)
            frame = NULL;
    } else {
        av_log(ctx, AV_LOG_ERROR, "frame has not been put to VPE encoder\n");
        return AVERROR_INVALIDDATA;
    }

    if (frame) {
        trans_frame = av_frame_alloc();
        if (!trans_frame)
            return AVERROR(ENOMEM);

        av_frame_move_ref(trans_frame, frame);
        vpi_frame         = (VpiFrm *)trans_frame->data[0];
        vpi_frame->pts    = trans_frame->pts;
        vpi_frame->dts    = trans_frame->pkt_dts;
        vpi_frame->valid  = 1;
        vpi_frame->opaque = (void *)trans_frame;
        if(ctx->dynamic_params_fp) {
            ama_encoder_get_dyn_params(ctx->dynamic_params_fp, ctx->sendFrameNum, &vpi_frame->dynamic_params);
        }
        snenc_queue_pts(ctx->pts_queue, vpi_frame->pts);

        //Dynamic IDR Insertion
        vpi_frame->key_frame = 0;
        /* Set frame to be encoded as IDR, if picture type is INTRA */
        if(trans_frame->pict_type == AV_PICTURE_TYPE_I) {
            av_log(ctx, AV_LOG_DEBUG, "dynamic idr ON for frame %d\n", ctx->sendFrameNum);
            vpi_frame->key_frame = 1;
        }

        av_log(ctx, AV_LOG_DEBUG, "line %d put frm %p, vpifrm %p\n", __LINE__, trans_frame, trans_frame->data[0]);
    } else {
        av_log(ctx, AV_LOG_DEBUG, "input image is empty, received EOF\n");
        ctx->eof_flush_en = 1;
        av_log(avctx, AV_LOG_DEBUG, "ENC Start Flush\n");
    }

    ret = vpi_encoder_put_frame(ctx->enc_handle, vpi_frame);
    if (ret) {
        av_log(ctx, AV_LOG_ERROR, "put frame to external encoder failed, error=%s\n",
                                       vpi_error_str(ret));
        return AVERROR_EXTERNAL;
    }

    //count sent frames
    if (vpi_frame) {
        ctx->sendFrameNum++;
        av_log(avctx, AV_LOG_DEBUG, "ENC Send Frame %d (%dx%d)\n",
                                     ctx->sendFrameNum,avctx->width,avctx->height);
    }
    return 0;
}

static int _sn_encode_receive_packet(AVCodecContext *avctx, AVPacket *avpkt)
{
    snEncContext *ctx = (snEncContext *)avctx->priv_data;
    AVFrame *frm;
    int pkt_size = 0;
    int retry_count = 1000000;
    int ret;
    int cmp;

    struct timeval cur_time;

    if (ctx->eof_flush_en == 0) {
        ret = sn_encode_send_frame(avctx);
        if (ret)
            return ret;
    }

    if (ctx->flush_done) {
        //return empty packet
        avpkt->data = NULL;
        avpkt->size = 0;

        gettimeofday(&cur_time, NULL);
        ctx->end_time = ((int64_t)(cur_time.tv_sec) *1000000) + cur_time.tv_usec;
        av_log(ctx, AV_LOG_DEBUG, "ctx->end_time = %ld\n", ctx->end_time);

        return AVERROR_EOF;
    }

    //allocate buffer to recv data.
    vpi_encoder_get_packet_size(ctx->enc_handle, &pkt_size);
    ret = av_new_packet(avpkt, pkt_size);
    if (ret) {
        av_log(avctx, AV_LOG_ERROR, "%s: av_new_packet error\n", __FUNCTION__);
        return ret;
    }

retry:
    ctx->vpi_pkt->buffer = avpkt->data;
    ctx->vpi_pkt->size   = avpkt->size;
    ret = vpi_encoder_get_packet(ctx->enc_handle, (void *)ctx->vpi_pkt);
    if (ret < 0) {
        av_log(avctx, AV_LOG_ERROR, "%s: vpi_encoder_get_packet failed\n", __FUNCTION__);
        return AVERROR_EXTERNAL;
    }

    //check for ULL mode exit
    if ((ctx->eof_flush_en)       &&
        (ctx->vpi_pkt->size == 0) &&
        (ctx->vpi_pkt->is_eof)) {
            //last packet has already been sent
            ctx->flush_done = 1;

            //return empty packet
            avpkt->data = NULL;
            avpkt->size = 0;

            gettimeofday(&cur_time, NULL);
            ctx->end_time = ((int64_t)(cur_time.tv_sec) *1000000) + cur_time.tv_usec;
            av_log(ctx, AV_LOG_DEBUG, "ctx->end_time = %ld\n", ctx->end_time);

            return AVERROR_EOF;
    }

    if (ctx->vpi_pkt->size == 0) {
        if (ctx->eof_flush_en) {
            av_log(avctx, AV_LOG_DEBUG, "ENC frame %d Recv No Data - Try Again (%d)\n", ctx->recvFrameNum, retry_count);
            --retry_count;
            if(retry_count) {
                ret = vpi_encoder_put_frame(ctx->enc_handle, NULL);
                if (ret < 0) {
                    av_log(avctx, AV_LOG_ERROR, "%s: vpi_encoder_put_frame failed, %s\n", __FUNCTION__, vpi_error_str(ret));
                    return AVERROR_EXTERNAL;
                }
                av_usleep(10000); //wait 10ms
                goto retry;
            } else {
                av_log(avctx, AV_LOG_ERROR, "%s: ENC Recv failed during flush\n",__FUNCTION__);
                return AVERROR_EXTERNAL;
            }
        } else  {
            av_packet_unref(avpkt);
            return AVERROR(EAGAIN); //send more data
        }
    }

    ctx->recvFrameNum++;
    av_log(avctx, AV_LOG_DEBUG, "ENC Recv Frame %d data=%d\n",ctx->recvFrameNum,  ctx->vpi_pkt->size);

    if(ctx->recvFrameNum == 1) {
        gettimeofday(&cur_time, NULL);
        ctx->start_time = ((int64_t)(cur_time.tv_sec) *1000000) + cur_time.tv_usec;
        av_log(ctx, AV_LOG_DEBUG, "ctx->start_time = %ld\n", ctx->start_time);
    }

    avpkt->size = ctx->vpi_pkt->size;
    avpkt->pts  = ctx->vpi_pkt->pts;
    if(ctx->vpi_pkt->key_frame){
        av_log(ctx, AV_LOG_DEBUG,"%u id keyframe\n", ctx->recvFrameNum);
        avpkt->flags |= AV_PKT_FLAG_KEY;
    }
    snenc_prepare_out_timestamp(avctx, avpkt);
    av_log(avctx, AV_LOG_DEBUG, "packet pts = %ld dts = %ld\n", avpkt->pts, avpkt->dts);

    if(ctx->enc_params.codec_format != (int)VPE_ENC_AV1_T2){
        for(cmp=0; cmp<ctx->nb_components; cmp++){
            ctx->sum_psnr[cmp] += ctx->vpi_pkt->psnr[cmp];
            ctx->sum_ssim[cmp] += ctx->vpi_pkt->ssim[cmp];
            ctx->max_psnr[cmp] = FFMAX(ctx->max_psnr[cmp], ctx->vpi_pkt->psnr[cmp]);
            ctx->max_ssim[cmp] = FFMAX(ctx->max_ssim[cmp], ctx->vpi_pkt->ssim[cmp]);
            ctx->min_psnr[cmp] = FFMIN(ctx->min_psnr[cmp], ctx->vpi_pkt->psnr[cmp]);
            ctx->min_ssim[cmp] = FFMIN(ctx->min_ssim[cmp], ctx->vpi_pkt->ssim[cmp]);
        }
    }

    //release encoded host frame
    frm = (AVFrame *)ctx->vpi_pkt->opaque;
    if(frm) {
        /* remove the device frame ref from avframe*/
        av_frame_free(&frm);
    }

    if(ctx->eof_flush_en) {
        if(ctx->vpi_pkt->is_eof) {
            ctx->flush_done = 1; //last pkt is being sent
        } else {
            ret = vpi_encoder_put_frame(ctx->enc_handle, NULL);
            if (ret < 0) {
                av_log(avctx, AV_LOG_ERROR, "%s: vpi_encoder_put_frame failed, %s\n", __FUNCTION__, vpi_error_str(ret));
                return AVERROR_EXTERNAL;
            }
        }
    }
    return 0;
}

static int sn_encode_receive_packet(AVCodecContext *avctx, AVPacket *avpkt)
{
    int ret;
    snEncContext *enc_ctx = avctx->priv_data;
    LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_ENC, enc_ctx->instance_name);
    ret = _sn_encode_receive_packet(avctx, avpkt);
    LOG_AMA_RESTORE_INSTANCE_INFO();
    return ret;
}

static av_cold int _sn_encode_close(AVCodecContext *avctx)
{
    snEncContext *ctx = (snEncContext *)avctx->priv_data;

#if CONFIG_LIBXRM
    xrm_enc_release(&ctx->xrm_ctx);
#endif

    //Log HW PSNR/SSIM values
    if(ctx->recvFrameNum > 1 && ctx->enc_params.codec_format != (int)VPE_ENC_AV1_T2) {
        int cmp;
        char comp[3][3];
        char *log_mean=NULL, *log_min=NULL, *log_max=NULL;
        log_mean = (char *)malloc(128*sizeof(char));
        if(!log_mean){
            av_log(avctx, AV_LOG_ERROR, "Failed to allocated memory for PSNR/SSIM logging");
            goto error_log;
        }
        log_min = (char *)malloc(128*sizeof(char));
        if(!log_min){
            av_log(avctx, AV_LOG_ERROR, "Failed to allocated memory for PSNR/SSIM logging");
            goto error_log;
        }
        log_max = (char *)malloc(128*sizeof(char));
        if(!log_max){
            av_log(avctx, AV_LOG_ERROR, "Failed to allocated memory for PSNR/SSIM logging");
            goto error_log;
        }

        if(ctx->nb_components == 3){
            strcpy(comp[0], " Y");
            strcpy(comp[1], " U");
            strcpy(comp[2], " V");
        } else {
            strcpy(comp[0], " Y");
            strcpy(comp[1], "UV");
            strcpy(comp[2], " -");
        }

        sprintf(log_mean, "PSNR MEAN ");
        sprintf(log_min, "PSNR MIN  ");
        sprintf(log_max, "PSNR MAX  ");
        for(cmp=0; cmp<ctx->nb_components; cmp++){
            sprintf(log_mean, "%s%s:%10.6f ", log_mean, comp[cmp], (ctx->sum_psnr[cmp]/ctx->recvFrameNum));
            sprintf(log_min, "%s%s:%10.6f ", log_min, comp[cmp], ctx->min_psnr[cmp]);
            sprintf(log_max, "%s%s:%10.6f ", log_max, comp[cmp], ctx->max_psnr[cmp]);
        }
        av_log(avctx, AV_LOG_INFO, "%s\n", log_mean);
        av_log(avctx, AV_LOG_INFO, "%s\n", log_min);
        av_log(avctx, AV_LOG_INFO, "%s\n", log_max);

        sprintf(log_mean, "SSIM MEAN ");
        sprintf(log_min, "SSIM MIN  ");
        sprintf(log_max, "SSIM MAX  ");
        for(cmp=0; cmp<ctx->nb_components; cmp++){
            sprintf(log_mean, "%s%s:%10.6f ", log_mean, comp[cmp], (ctx->sum_ssim[cmp]/ctx->recvFrameNum));
            sprintf(log_min, "%s%s:%10.6f ", log_min, comp[cmp], ctx->min_ssim[cmp]);
            sprintf(log_max, "%s%s:%10.6f ", log_max, comp[cmp], ctx->max_ssim[cmp]);
        }
        av_log(avctx, AV_LOG_INFO, "%s\n", log_mean);
        av_log(avctx, AV_LOG_INFO, "%s\n", log_min);
        av_log(avctx, AV_LOG_INFO, "%s\n", log_max);

    //In case of error skip printing psnr/ssim values
    //but continue with uninit
    error_log:
        if(log_mean){
            free(log_mean);
        }
        if(log_min){
            free(log_min);
        }
        if(log_max){
            free(log_max);
        }
    }

    if (ctx->dynamic_params_fp) {
        fclose(ctx->dynamic_params_fp);
    }
    vpi_encoder_close(ctx->enc_handle);
    av_frame_free(&ctx->frame);

    if(ctx->enc_params.level){
        free(ctx->enc_params.level);
        ctx->enc_params.level = NULL;
    }
    if(ctx->enc_params.profile){
        free(ctx->enc_params.profile);
        ctx->enc_params.profile = NULL;
    }

    if (ctx->vpi_pkt) {
        vpi_memory_free_struct(ctx->vpi_pkt);
    }
    av_buffer_unref(&ctx->hwframe);
    av_buffer_unref(&ctx->hwdevice);

    av_fifo_freep2(&ctx->pts_queue);

    return 0;
}

static av_cold int sn_encode_close(AVCodecContext *avctx)
{
    int ret;
    snEncContext *enc_ctx = avctx->priv_data;
    LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_ENC, enc_ctx->instance_name);
    ret = _sn_encode_close(avctx);
    LOG_AMA_RESTORE_INSTANCE_INFO();
    return ret;
}



static const FFCodecDefault snenc_h264_defaults[] = {
    {"b", "0"},
    {"g", "-1"},
    { NULL },
};

static const FFCodecDefault snenc_hevc_defaults[] = {
    {"b", "0"},
    {"g", "-1"},
    { NULL },
};

static const FFCodecDefault snenc_av1_defaults[] = {
    {"b", "0"},
    {"g", "-1"},
    { NULL },
};

static const AVClass snenc_h264_class = {
    .class_name = "h264e_ama",
    .item_name  = av_default_item_name,
    .option     = h264enc_options,
    .version    = LIBAVUTIL_VERSION_INT,
};

static const AVClass snenc_hevc_class = {
    .class_name = "hevce_ama",
    .item_name  = av_default_item_name,
    .option     = hevcenc_options,
    .version    = LIBAVUTIL_VERSION_INT,
};

static const AVClass snenc_av1_class = {
    .class_name = "av1_ama",
    .item_name  = av_default_item_name,
    .option     = av1enc_options,
    .version    = LIBAVUTIL_VERSION_INT,
};

static const AVCodecHWConfigInternal *hw_configs[] =
    { &(const AVCodecHWConfigInternal) {
        .public = {
            .pix_fmt = AV_PIX_FMT_VPE,
            .methods = AV_CODEC_HW_CONFIG_METHOD_HW_DEVICE_CTX |
                        AV_CODEC_HW_CONFIG_METHOD_HW_FRAMES_CTX,
            .device_type = AV_HWDEVICE_TYPE_AMA,
        },
       .hwaccel = NULL,
      },
      NULL };

FFCodec ff_snh264_vpe_encoder = {
    .p.name           = "h264_ama",
    .p.long_name      = NULL_IF_CONFIG_SMALL("H264"),
    .p.type           = AVMEDIA_TYPE_VIDEO,
    .p.id             = AV_CODEC_ID_H264,
    .priv_data_size = sizeof(snEncContext),
    .init           = sn_encode_init,
    FF_CODEC_RECEIVE_PACKET_CB(sn_encode_receive_packet),
    .close          = sn_encode_close,
    .p.priv_class     = &snenc_h264_class,
    .p.capabilities   = AV_CODEC_CAP_DELAY | AV_CODEC_CAP_HARDWARE,
    .defaults       = snenc_h264_defaults,
    .p.pix_fmts       = (const enum AVPixelFormat[]) {
        AV_PIX_FMT_VPE,
        AV_PIX_FMT_NONE
  	},
    .hw_configs     = hw_configs,
    .p.wrapper_name   = "ama",
};

FFCodec ff_snhevc_vpe_encoder = {
    .p.name           = "hevc_ama",
    .p.long_name      = NULL_IF_CONFIG_SMALL("HEVC"),
    .p.type           = AVMEDIA_TYPE_VIDEO,
    .p.id             = AV_CODEC_ID_HEVC,
    .priv_data_size = sizeof(snEncContext),
    .init           = sn_encode_init,
    FF_CODEC_RECEIVE_PACKET_CB(sn_encode_receive_packet),
    .close          = sn_encode_close,
    .p.priv_class     = &snenc_hevc_class,
    .p.capabilities   = AV_CODEC_CAP_DELAY | AV_CODEC_CAP_HARDWARE,
    .defaults       = snenc_hevc_defaults,
    .p.pix_fmts       = (const enum AVPixelFormat[]) {
        AV_PIX_FMT_VPE,
        AV_PIX_FMT_NONE
  	},
    .hw_configs     = hw_configs,
    .p.wrapper_name   = "ama",
};

FFCodec ff_snav1_vpe_encoder = {
    .p.name             = "av1_ama",
    .p.long_name        = NULL_IF_CONFIG_SMALL("AV1"),
    .p.type             = AVMEDIA_TYPE_VIDEO,
    .p.id               = AV_CODEC_ID_AV1,
    .priv_data_size   = sizeof(snEncContext),
    .init             = sn_encode_init,
    FF_CODEC_RECEIVE_PACKET_CB(sn_encode_receive_packet),
    .close            = sn_encode_close,
    .p.priv_class       = &snenc_av1_class,
    .p.capabilities     = AV_CODEC_CAP_DELAY | AV_CODEC_CAP_HARDWARE,
    .defaults         = snenc_av1_defaults,
    .p.pix_fmts         = (const enum AVPixelFormat[]) {
        AV_PIX_FMT_VPE,
        AV_PIX_FMT_NONE
    },
    .hw_configs     = hw_configs,
    .p.wrapper_name     = "ama",
};
