/*
* AMD ROI side data overlay test filter
*
* This file is part of FFmpeg.
*
* FFmpeg is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* FFmpeg is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with FFmpeg; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */


#include "avfilter.h"
#include "filters.h"
#include "libavutil/opt.h"
#include "libavutil/pixdesc.h"
#include "libavutil/hwcontext.h"
#include "log_ama.h"


struct ROIOverlayContext;

typedef struct ROIOverlayContext {
    const AVClass *class;
    AVBufferRef       *hwframes_ref;
    int skip_frames_val;
    int highlight_threshold_val;
    int frame_cnt;
    char instance_name[64];
} ROIOverlayContext;

static const enum AVPixelFormat pix_fmts[] = {
    AV_PIX_FMT_YUV420P,
    AV_PIX_FMT_NONE
};


static av_cold int init(AVFilterContext *ctx)
{
    ROIOverlayContext *s = ctx->priv;
    char *ptr = strrchr(ctx->name, '_');
    snprintf(s->instance_name, sizeof(s->instance_name), (ptr) ? "roi_overlay_ama%s" : "roi_overlay_ama@%p", (ptr) ? ptr : s);

    return 0;
}

static int roi_overlay_filter_frame(AVFilterLink *inlink, AVFrame *frame)
{
    AVFilterContext *ctx = inlink->dst;
    ROIOverlayContext *s = ctx->priv;
    AVHWFramesContext *hwframe_ctx;
    AVBufferRef *hw_frames_ctx;
    int err, width, height;
    AVDictionaryEntry *t = NULL;
    AVFrame *output = NULL;
    AVFrameSideData * av_roi_sidedata = NULL;

    output = av_frame_clone(frame);
    if (!output) {
        err = AVERROR(ENOMEM);
    }

    av_roi_sidedata = av_frame_get_side_data(frame,
                                        AV_FRAME_DATA_SB_ML_ROI);

    if (av_roi_sidedata) {
        unsigned char *detected_roi = av_roi_sidedata->data;

        t = av_dict_get(av_roi_sidedata->metadata, "width", NULL, 0);
        if (t)
            width = atoi(t->value);
        t = av_dict_get(av_roi_sidedata->metadata, "height", NULL, 0);
        if (t)
            height = atoi(t->value);

        s->frame_cnt++;
        // loop for the roi dimension
        for (int ry=0; ry < height; ry++) {
            for (int rx=0; rx < width; rx++) {
                int index = (ry * width) + rx;

                // if valid activation, loop for 16x16 box drawing
                if ((detected_roi[index] >= s->highlight_threshold_val)) {
                    int locy=ry*16, locx=rx*16;
                    for(int y = locy; y < locy+16; y++) {
                        for(int x = locx; x < locx+16; x++) {

                            // ensure to draw within frame dimension
                            if ((x < output->width) && (y < output->height)) {
                                int offsetY = 1 * ((x) + (y * output->linesize[0]));

                                // use skip frame as per configuraion via filter parameter
                                if ((!s->skip_frames_val) || ((s->frame_cnt/s->skip_frames_val) & 0x01)) {
                                    int temp = output->data[0][offsetY + 0] + ((detected_roi[index]*detected_roi[index])/(3*255));
                                    if (temp > 255)
                                        temp = 255;
                                    output->data[0][offsetY + 0] = temp;
                                }
                            }
                        }
                    }
                }
            }
        }
        av_frame_remove_side_data(output, AV_FRAME_DATA_SB_ML_ROI);
    }

    av_frame_free(&frame);
    return ff_filter_frame(inlink->dst->outputs[0], output);
}

static int filter_frame(AVFilterLink *inlink, AVFrame *frame)
{
    int ret;
    AVFilterContext *ctx = inlink->dst;
    ROIOverlayContext *s = ctx->priv;

    LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_GENERAL, s->instance_name);
    ret = roi_overlay_filter_frame(inlink, frame);
    LOG_AMA_RESTORE_INSTANCE_INFO();

    return ret;
}


#define OFFSET(x) offsetof(ROIOverlayContext, x)
#define FLAGS (AV_OPT_FLAG_VIDEO_PARAM | AV_OPT_FLAG_FILTERING_PARAM)

static const AVOption roi_overlay_options[] = {
    { "skip_frames",          "set the frames to skip for overlay highlight",   OFFSET(skip_frames_val), AV_OPT_TYPE_INT, {.i64=0}, 0, 255, FLAGS },
    { "highlight_threshold",  "set the threshold value for overlay highlight",  OFFSET(highlight_threshold_val), AV_OPT_TYPE_INT, {.i64=128}, 0, 255, FLAGS },
    { NULL }
};


AVFILTER_DEFINE_CLASS(roi_overlay);

static const AVFilterPad roi_overlay_inputs[] = {
    {
        .name           = "default",
        .type           = AVMEDIA_TYPE_VIDEO,
        .filter_frame   = filter_frame,
    },
};

static const AVFilterPad roi_overlay_outputs[] = {
    {
        .name = "default",
        .type = AVMEDIA_TYPE_VIDEO,
    },
};

const AVFilter ff_vf_roi_overlay = {
    .name          = "roi_overlay_ama",
    .description   = NULL_IF_CONFIG_SMALL("Draw the visual highlight of roi detection on frame."),
    .priv_size     = sizeof(ROIOverlayContext),
    .priv_class    = &roi_overlay_class,
    .nb_inputs     = 1,
    .nb_outputs    = 1,
    .init          = init,
    .flags         = AVFILTER_FLAG_SUPPORT_TIMELINE_GENERIC,
    FILTER_INPUTS(roi_overlay_inputs),
    FILTER_OUTPUTS(roi_overlay_outputs),
    FILTER_PIXFMTS_ARRAY(pix_fmts),
};
