/*
 * VeriSilicon VPE VIP NBG process interface
 * Copyright (C) 2020 VeriSilicon Holdings Co., Ltd.
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file
 * Filter implementing * filter using deep convolutional networks by nbg.
 * http://openaccess.thecvf.com/content_ECCV_2018/html/Xia_Li_Recurrent_Squeeze-and-Excitation_Context_ECCV_2018_paper.html
 */

#include "libavutil/avstring.h"
#include "libavformat/avio.h"
#include "libavutil/buffer.h"
#include "libavutil/buffer_internal.h"
#include "libavutil/opt.h"
#include "libavutil/pixdesc.h"
#include "libavutil/hwcontext.h"
#include "libavutil/hwcontext_vpe.h"
#include "libavutil/imgutils.h"
#include "libavutil/parseutils.h"

#include <time.h>
#if defined(__linux__)
#include <sys/time.h>
#elif defined(_WIN32)
#include <winsock.h>
#endif

#include "vpi_types.h"
#include "vpi_api.h"
#include "avfilter.h"
#include "formats.h"
#include "internal.h"
#include "filters.h"
#include "framesync.h"
#include "log_ama.h"
#if CONFIG_LIBXRM
#include "xrm_ml_interface.h"
#endif
static av_cold void uninit(AVFilterContext *ctx);

struct timeval cur_time;

typedef struct VPEVIPNBGContext {
    const AVClass *class;
    FFFrameSync fs;
    const AVPixFmtDescriptor *desc;
    AVFrame **frames;
    AVBufferRef *in_hwframe_ctx;
    AVBufferRef *out_hwframe_ctx;
    AVBufferRef *hwdevice;

#if CONFIG_LIBXRM
    XrmMLContext xrm_ml_ctx;
#endif

    VpiProcInitParams vpe_vip_nbg_params;

    char *vpe_vip_nbg_extra_params;

    /*logical device id*/
    int logical_dev;

    /*setting input channel number*/
    int nb_inputs;

    AVDictionary *dict;
    char *vpe_vip_extra_params;

    VpiProcHandle vpe_vip_nbg_handle;
    VpiChannelInfo ch_info[VPE_CHANNEL_NUM];

    /* add for perf test */
    int64_t  start_time;
    int64_t  end_time;
    int64_t  mid_time;

    char *osml_model;
    char *osml_model_args;


    char instance_name[64];
} VPEVIPNBGContext;

#define CLIP(x, min, max) (x < min ? min : (x > max ? max : x))
#define OFFSET(x) offsetof(VPEVIPNBGContext, x)
#define FLAGS AV_OPT_FLAG_VIDEO_PARAM | AV_OPT_FLAG_FILTERING_PARAM
static const AVOption ml_ama_options[] = {
                                                    { "model",
                                                    "Model to excecute",
                                                    OFFSET(osml_model),
                                                    AV_OPT_TYPE_STRING,
                                                    { .str = NULL },
                                                    0,
                                                    0,
                                                    FLAGS },
                                                    { "model_args",
                                                    "model specific arguments",
                                                    OFFSET(osml_model_args),
                                                    AV_OPT_TYPE_STRING,
                                                    { .str = NULL },
                                                    0,
                                                    0,
                                                    FLAGS },
                                                {
                                                    "core_id",
                                                    "set logical device id",
                                                    OFFSET(logical_dev),
                                                    AV_OPT_TYPE_INT,
                                                    { .i64 = -1 },
                                                    -1,
                                                    3,
                                                    .flags = FLAGS },
                                                {
                                                    "nb_inputs",
                                                    "set number of inputs",
                                                    OFFSET(nb_inputs),
                                                    AV_OPT_TYPE_INT,
                                                    { .i64 = 1 },
                                                    1,
                                                    8,
                                                    .flags = FLAGS },
                                                { 
                                                    "expert_options",
                                                    "set vip extra params",
                                                    OFFSET(vpe_vip_extra_params),
                                                    AV_OPT_TYPE_STRING,
                                                    { .str = NULL },
                                                    .flags = FLAGS },
                                                { NULL },
                                                };

AVFILTER_DEFINE_CLASS(ml_ama);

static int vpe_vip_nbg_init_hwdev(AVFilterContext *ctx)
{
    VPEVIPNBGContext *vip_nbg_ctx = ctx->priv;

    av_log(ctx, AV_LOG_DEBUG, "%s(%d)\n", __FUNCTION__, __LINE__);

    if (!vip_nbg_ctx->hwdevice) {
        if (ctx->hw_device_ctx) {
            vip_nbg_ctx->hwdevice = av_buffer_ref(ctx->hw_device_ctx);
            if (!vip_nbg_ctx->hwdevice) {
                return AVERROR(ENOMEM);
            }
        } else {
            return AVERROR_UNKNOWN;
        }
    }
    return 0;
}

static int vpe_vip_nbg_query_formats(AVFilterContext *ctx)
{
    VPEVIPNBGContext *vip_nbg_ctx      = ctx->priv;
    AVVpeHWConfig *hwconfig            = NULL;
    AVHWFramesConstraints *constraints = NULL;
    AVFilterFormats *in_fmts;
    AVFilterFormats *out_fmts;
    int ret;

    ret = vpe_vip_nbg_init_hwdev(ctx);
    if (ret < 0) {
        goto error;
    }

    hwconfig = av_hwdevice_hwconfig_alloc(vip_nbg_ctx->hwdevice);
    if (!hwconfig) {
        ret = AVERROR(ENOMEM);
        goto error;
    }

    hwconfig->id = VPE_HW_VIP_NBG;
    av_log(ctx, AV_LOG_DEBUG, "hwconfig->id=%d\n", hwconfig->id);

    constraints = av_hwdevice_get_hwframe_constraints(vip_nbg_ctx->hwdevice, hwconfig);
    if (!constraints) {
        av_log(ctx, AV_LOG_ERROR, "failed to get hw frame constraints\n");
        ret = AVERROR(ENOMEM);
        goto error;
    }

    out_fmts = ff_make_format_list(constraints->valid_hw_formats);
    ret      = ff_formats_ref(out_fmts, &ctx->outputs[0]->incfg.formats);
    if (ret < 0)
        goto error;

    in_fmts = ff_make_format_list(constraints->valid_hw_formats);
    ret = ff_formats_ref(in_fmts, &ctx->inputs[0]->outcfg.formats);
    if (ret < 0)
        goto error;

    av_hwframe_constraints_free(&constraints);
    av_freep(&hwconfig);

    return 0;

error:
    av_log(ctx, AV_LOG_ERROR, "%s failed\n", __FUNCTION__);

    av_hwframe_constraints_free(&constraints);
    av_freep(&hwconfig);

    return ret;
}

static void set_instance_name(VPEVIPNBGContext *s, const char* ffmpeg_name)
{
    char *ptr = strrchr(ffmpeg_name, '_');
    snprintf(s->instance_name, sizeof(s->instance_name), (ptr) ? "ml_ama%s" : "ml_ama@%p", (ptr) ? ptr : s);
}

static int query_formats(AVFilterContext *ctx)
{
    int ret;
    VPEVIPNBGContext *s = ctx->priv;

    LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_ML, s->instance_name);
    ret = vpe_vip_nbg_query_formats(ctx);
    LOG_AMA_RESTORE_INSTANCE_INFO();

    return ret;
}

static int vpe_vip_nbg_check_constraints(AVFilterContext *ctx)
{
    VPEVIPNBGContext *vip_nbg_ctx      = ctx->priv;
    AVFilterLink *inlink               = ctx->inputs[0];
    AVVpeHWConfig *hwconfig            = NULL;
    AVHWFramesConstraints *constraints;
    int ret = 0;

    hwconfig = av_hwdevice_hwconfig_alloc(vip_nbg_ctx->hwdevice);
    if (!hwconfig) {
        ret = AVERROR(ENOMEM);
        goto error;
    }

    hwconfig->id = VPE_HW_VIP_NBG;
    av_log(ctx, AV_LOG_DEBUG, "hwconfig->id = %d\n", hwconfig->id);

    constraints = av_hwdevice_get_hwframe_constraints(vip_nbg_ctx->hwdevice, hwconfig);
    if (!constraints) {
        ret = AVERROR(ENOMEM);
        goto error;
    }

    av_log(ctx, AV_LOG_DEBUG, "get constraints %d,%d,%d,%d\n",
           constraints->min_width, constraints->min_height,
           constraints->max_width, constraints->max_height);

    if (inlink->w < constraints->min_width ||
        inlink->h < constraints->min_height ||
        inlink->w > constraints->max_width ||
        inlink->h > constraints->max_height) {
        av_log(ctx, AV_LOG_ERROR,
               "Hardware does not support image "
               "size %dx%d (constraints: width %d-%d height %d-%d).\n",
               inlink->w, inlink->h, constraints->min_width,
               constraints->max_width, constraints->min_height,
               constraints->max_height);
        ret = AVERROR(EINVAL);
        goto error;
    }

    return ret;

error:
    av_log(ctx, AV_LOG_ERROR, "%s failed\n", __FUNCTION__);

    if (constraints)
        av_hwframe_constraints_free(&constraints);
    if (hwconfig)
        av_freep(&hwconfig);

    return ret;
}

static int vpe_vip_nbg_init_hwctx(AVFilterContext *ctx)
{
    VPEVIPNBGContext *vpe_vip_nbg_ctx = ctx->priv;
    VpiProcInitParams *param = &vpe_vip_nbg_ctx->vpe_vip_nbg_params;
    AVHWFramesContext *hw_frame_ctx;
    AVHWDeviceContext *device_ctx;
    AVVpeDeviceContext *device_hwctx;
    AVVpeFramesContext *frame_hwctx;
    AVFilterLink *inlink = ctx->inputs[0];
    int ret = 0;

    av_log(ctx, AV_LOG_DEBUG, "%s(%d) param addr=%p\n", __FUNCTION__, __LINE__, param);

    if (inlink->hw_frames_ctx) {
        vpe_vip_nbg_ctx->in_hwframe_ctx = av_buffer_ref(inlink->hw_frames_ctx);
        if (!vpe_vip_nbg_ctx->in_hwframe_ctx) {
            ret = AVERROR(ENOMEM);
            return ret;
        }

        hw_frame_ctx = (AVHWFramesContext *)vpe_vip_nbg_ctx->in_hwframe_ctx->data;
    } else {
        ret = AVERROR_UNKNOWN;
        return ret;
    }

    ret = vpe_vip_nbg_check_constraints(ctx);
    if(ret < 0) {
        av_log(ctx, AV_LOG_ERROR, "vpe_vip_nbg_check_constraints failed\n");
        return ret;
    }

    device_ctx = hw_frame_ctx->device_ctx;
    device_hwctx = device_ctx->hwctx;
    frame_hwctx = hw_frame_ctx->hwctx;

    vpi_device_set_ctx(param, device_hwctx->vpe_dev_ctx, VPE_HW_VIP_NBG);
    param->walign = frame_hwctx->proc_info->ch_info[0].walign;
    param->halign = frame_hwctx->proc_info->ch_info[0].halign;
    av_log(ctx, AV_LOG_DEBUG, "%s: vpe_mem_ctx @%p\n", __func__,
           frame_hwctx->vpe_mem_ctx);
    av_log(ctx, AV_LOG_DEBUG, "%s: vpe_dev_ctx @%p\n", __func__,
            device_hwctx->vpe_dev_ctx);

    return 0;
}

static int vpe_vip_nbg_process_frame(FFFrameSync *fs)
{
    AVFilterContext *ctx  = fs->parent;
    AVFilterLink *outlink = ctx->outputs[0];
    VPEVIPNBGContext *s   = fs->opaque;
    AVFrame **frame_in    = s->frames;
    AVFrame *frame_out    = NULL;
    VpiFrm *out_pic       = NULL;
    VpiFrm *in_pic        = NULL;
    VpiFrm *in_temp       = NULL;
    VpiFrm *next_pic      = NULL;
    int ret               = 0;
    int i                 = 0;

    for (i = 0; i < s->nb_inputs; i++) {
        if ((ret = ff_framesync_get_frame(&s->fs, i, &frame_in[i], 0)) < 0)
            return ret;

        in_temp = (VpiFrm *)(frame_in[i]->data[0]);
        if (!in_temp)
            return AVERROR_INVALIDDATA;

        if (i == 0) {
            in_pic   = in_temp;
            next_pic = in_pic;
        } else {
            next_pic->next = in_temp;
            next_pic       = next_pic->next;
        }
    }

    if (!in_pic)
        return AVERROR_INVALIDDATA;

    frame_out = ff_get_video_buffer(outlink, outlink->w, outlink->h);
    if (!frame_out)
        return AVERROR(ENOMEM);

    ret = av_frame_copy_props(frame_out, frame_in[0]);
    if (ret < 0){
        av_log(ctx, AV_LOG_ERROR, "av_frame_copy_props failed\n");
        return AVERROR_EXTERNAL;
    }

    frame_out->pts = av_rescale_q(s->fs.pts, s->fs.time_base, outlink->time_base);
    frame_out->sample_aspect_ratio = outlink->sample_aspect_ratio;

    out_pic = (VpiFrm *)(frame_out->data[0]);
    if (!out_pic){
        return AVERROR_INVALIDDATA;
    }

    ret = vpi_processor_process_frame(s->vpe_vip_nbg_handle, in_pic, out_pic);
    if (ret < 0) {
        av_log(ctx, AV_LOG_ERROR, "vpi_processor_process_frame failed, error=%s\n", vpi_error_str(ret));
        return AVERROR_EXTERNAL;
    }

    frame_out->format      = AV_PIX_FMT_VPE;
    frame_out->width       = out_pic->ch_info[0].width;
    frame_out->height      = out_pic->ch_info[0].height;

    out_pic->log_frm_num = in_pic->log_frm_num;
    vpi_frame_unref(in_pic);

    ret = ff_filter_frame(outlink, frame_out);
    return ret;
}

static int vpe_vip_nbg_config_params(AVFilterLink *outlink)
{
    AVFilterContext *ctx = outlink->src;
    VPEVIPNBGContext *vip_nbg_ctx = ctx->priv;
    AVHWFramesContext *hwframe_ctx = NULL;
    AVVpeFramesContext *vpeframe_ctx = NULL;
    const int outlink_idx  = FF_OUTLINK_IDX(outlink);
    FFFrameSyncIn *in = NULL;
    VpiProcInitParams *params = &vip_nbg_ctx->vpe_vip_nbg_params;
    int i = 0, ret = 0;
    AVFilterLink *inlink = NULL;
    VpiProcInfo info;
    AVDictionaryEntry *en = NULL;
    AVHWFramesContext *fmctx =  NULL;
#if CONFIG_LIBXRM
    XrmInterfaceProperties xrm_props;
    int dev_index = -1;
    int xrm_status = 0;
    memset(&xrm_props, 0, sizeof(xrm_props));
#endif

    if (!ctx->inputs[0])
        return AVERROR(EINVAL);

    //config input
    if ((ret = ff_framesync_init(&vip_nbg_ctx->fs, ctx, vip_nbg_ctx->nb_inputs)) < 0)
        return ret;

    in = vip_nbg_ctx->fs.in;
    vip_nbg_ctx->fs.opaque = vip_nbg_ctx;
    vip_nbg_ctx->fs.on_event = vpe_vip_nbg_process_frame;

    for (i = 0; i < vip_nbg_ctx->nb_inputs; i++) {
        inlink = ctx->inputs[i];

        in[i].time_base = inlink->time_base;
        in[i].sync   = 1;
        in[i].before = EXT_STOP;
        in[i].after  = EXT_INFINITY;
    }
    inlink = ctx->inputs[0];

    if (ff_framesync_configure(&vip_nbg_ctx->fs) < 0) {
        av_log(ctx, AV_LOG_ERROR, "failed to configure ff_framesync.\n");
        return AVERROR(EINVAL);
    }
    outlink->time_base = vip_nbg_ctx->fs.time_base;

    ret = vpe_vip_nbg_init_hwctx(ctx);
    if (ret < 0){
        av_log(ctx, AV_LOG_DEBUG, "VPE vip_nbg_hwctx failed.\n");
        return AVERROR(ENOMEM);
    }

    outlink->hw_frames_ctx = av_hwframe_ctx_alloc(vip_nbg_ctx->hwdevice);
    if (!outlink->hw_frames_ctx) {
        av_log(ctx, AV_LOG_ERROR, "hw_frames_ctx alloc failed\n");
        return AVERROR(ENOMEM);
    }

    hwframe_ctx = (AVHWFramesContext *) outlink->hw_frames_ctx->data;
    if (!hwframe_ctx->pool) {
        AVHWFramesContext*  fmctx = (AVHWFramesContext*) (ctx->inputs[0]->hw_frames_ctx->data);

        hwframe_ctx->width = inlink->w;
        hwframe_ctx->height = inlink->h;
        hwframe_ctx->format = AV_PIX_FMT_VPE;
        hwframe_ctx->sw_format = fmctx->sw_format;

        if ((ret = av_hwframe_ctx_init(outlink->hw_frames_ctx)) < 0) {
            av_log(ctx, AV_LOG_ERROR, "av_hwframe_ctx_init for output failed\n");
            return AVERROR(EINVAL);
        }
    }

    vip_nbg_ctx->out_hwframe_ctx = av_buffer_ref(outlink->hw_frames_ctx);

    if (ctx->inputs[0]->format == AV_PIX_FMT_VPE) {
        fmctx = (AVHWFramesContext *)(ctx->inputs[0]->hw_frames_ctx->data);
        params->input_sw_format = av_get_pix_fmt_name(fmctx->sw_format);
        params->input_format    = av_get_pix_fmt_name(ctx->inputs[0]->format);
        av_log(ctx, AV_LOG_DEBUG,
               "%s(%d), fmctx->sw_format=%d, params->input_sw_format=%s.\n",
               __func__, __LINE__, fmctx->sw_format, params->input_sw_format);
    } else {
        av_log(ctx, AV_LOG_ERROR, "%s: Format: %s not supported.\n", __func__,
               av_get_pix_fmt_name(ctx->inputs[0]->format));
        return AVERROR(EINVAL);
    }

    hwframe_ctx  = (AVHWFramesContext *)outlink->hw_frames_ctx->data;
    vpeframe_ctx = hwframe_ctx->hwctx;
    if (vpeframe_ctx == NULL) {
        ret = AVERROR_EXTERNAL;
        goto error;
    }

    params->mem_ctx_in = vpeframe_ctx->vpe_mem_ctx;
    params->logical_dev = vip_nbg_ctx->logical_dev; //set default -1 if not set
    params->nb_frames = vip_nbg_ctx->nb_inputs;
    params->width = ctx->inputs[0]->w;
    params->height = ctx->inputs[0]->h;
    params->processor_type = VPE_PROCESSOR_VIP_NBG;
    av_log(ctx, AV_LOG_DEBUG, "params init: nb_frames=%d, width=%d, height=%d\n",
                params->nb_frames, params->width, params->height);

    hwframe_ctx         = (AVHWFramesContext *)vip_nbg_ctx->out_hwframe_ctx->data;
    hwframe_ctx->format = ctx->inputs[0]->format;
    hwframe_ctx->width  = inlink->w;
    hwframe_ctx->height = inlink->h;
    //default use input's format
    hwframe_ctx->sw_format = fmctx->sw_format;


    //configure output, default use input[0] as output
    outlink->w                   = ctx->inputs[0]->w;
    outlink->h                   = ctx->inputs[0]->h;
    outlink->frame_rate          = ctx->inputs[0]->frame_rate;
    outlink->sample_aspect_ratio = ctx->inputs[0]->sample_aspect_ratio;
    outlink->time_base           = vip_nbg_ctx->fs.time_base;
    outlink->format              = hwframe_ctx->sw_format;

    ret = av_hwframe_ctx_init(outlink->hw_frames_ctx);
    if (ret < 0) {
        av_log(ctx, AV_LOG_ERROR, "%s(%d): the frame ctx init fail(%d).\n", __FUNCTION__, __LINE__, ret);
        goto error;
    }

#if CONFIG_LIBXRM
    // Reserve xrm resource
    xrm_props.width = vip_nbg_ctx->vpe_vip_nbg_params.width;
    xrm_props.height = vip_nbg_ctx->vpe_vip_nbg_params.height;
    xrm_props.fps_num = inlink->frame_rate.num;
    xrm_props.fps_den = inlink->frame_rate.den;
    vip_nbg_ctx->xrm_ml_ctx.ml_core_id = vip_nbg_ctx->logical_dev;
    strncpy(xrm_props.model, vip_nbg_ctx->osml_model, sizeof(xrm_props.model)-1);
    av_log(ctx, AV_LOG_DEBUG,"osml_model = %s\n", xrm_props.model);
    if(vip_nbg_ctx->osml_model_args!=NULL){
        strncpy(xrm_props.model_args, vip_nbg_ctx->osml_model_args, sizeof(xrm_props.model_args)-1);
        av_log(ctx, AV_LOG_DEBUG,"osml_model_args = %s\n", xrm_props.model_args);
    }
    if(vpi_device_get_device_index(vip_nbg_ctx->vpe_vip_nbg_params.hw_ctx, &dev_index) < 0) {
        return AVERROR(EINVAL);
    }
    xrm_status = xrm_ml_reserve(&vip_nbg_ctx->xrm_ml_ctx, dev_index, &xrm_props);
    if (xrm_status == XRM_ERROR) {
        return AVERROR(EBUSY);
    }else if(xrm_status == XRM_SUCCESS) {
        params->logical_dev = vip_nbg_ctx->xrm_ml_ctx.ml_core_id;
    }
    av_log(ctx, AV_LOG_DEBUG,"ml_core_id = %d\n", params->logical_dev);
#endif

    //config process
    ret = vpi_processor_init(params, &vip_nbg_ctx->vpe_vip_nbg_handle);
    if (ret < 0){
        av_log(ctx, AV_LOG_ERROR, "%s(%d): VPI VIP processor init failed: %s\n", __FUNCTION__, __LINE__, vpi_error_str(ret));
        return AVERROR_EXTERNAL;
    }

    while((en = av_dict_get(vip_nbg_ctx->dict, "", en, AV_DICT_IGNORE_SUFFIX))) {
        if(vpi_processor_set_param(vip_nbg_ctx->vpe_vip_nbg_handle, en->key, en->value) < 0)
            av_log(ctx, AV_LOG_WARNING,
                   "Error parsing option '%s = %s'. Please check if it is valid for vip9000 options.\n",
                   en->key, en->value);
    }

    ret = vpi_processor_open(vip_nbg_ctx->vpe_vip_nbg_handle);
    if (ret < 0) {
        av_log(ctx, AV_LOG_ERROR, "%s(%d): VPI VIP processor open failed: %s.\n", __FUNCTION__, __LINE__, vpi_error_str(ret));
        goto error;
    }

    //extract chanel info
    ret = vpi_processor_get_info(vip_nbg_ctx->vpe_vip_nbg_handle, &info);
    if(ret < 0) {
        av_log(ctx, AV_LOG_ERROR, "%s(%d): get processor info fail(%d).\n", __FUNCTION__, __LINE__, ret);
        goto error;
    }
    memcpy(vip_nbg_ctx->ch_info, &info.ch_info, sizeof(VpiChannelInfo) * VPE_CHANNEL_NUM);

    if (vpeframe_ctx->proc_info == NULL) {
        av_log(ctx, AV_LOG_ERROR, "output frame proc info is empty\n");
        ret = AVERROR_EXTERNAL;
        goto error;
    }
    vpeframe_ctx->proc_info->ch_info[outlink_idx] = vip_nbg_ctx->ch_info[outlink_idx];

    ret = vpi_memory_set_map_info(vpeframe_ctx->vpe_mem_ctx, 0, 0);
    if (ret < 0) {
        av_log(ctx, AV_LOG_ERROR, "vpi_memory_set_map_info failed\n");
        goto error;
    }

    ret = vpi_memory_set_single_channel_info(vpeframe_ctx->vpe_mem_ctx, 0,
                                             &vip_nbg_ctx->ch_info[outlink_idx],
                                             VPE_HW_VIP_NBG);
    if (ret < 0) {
        av_log(ctx, AV_LOG_ERROR,
               "vpi_memory_set_single_channel_info failed\n");
        goto error;
    }

    ctx->outputs[0]->format = AV_PIX_FMT_VPE;

error:
    return ret;
}

static int config_params(AVFilterLink *outlink)
{
    int ret;
    AVFilterContext *ctx = outlink->src;
    VPEVIPNBGContext *s = ctx->priv;

    LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_ML, s->instance_name);
    ret = vpe_vip_nbg_config_params(outlink);
    LOG_AMA_RESTORE_INSTANCE_INFO();

    return ret;
}

static int vpe_vip_nbg_activate(AVFilterContext *ctx)
{
    VPEVIPNBGContext *vip_nbg_ctx = ctx->priv;

    return ff_framesync_activate(&vip_nbg_ctx->fs);
}

static int activate(AVFilterContext *ctx)
{
    int ret;
    VPEVIPNBGContext *s = ctx->priv;

    LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_ML, s->instance_name);
    ret = vpe_vip_nbg_activate(ctx);
    LOG_AMA_RESTORE_INSTANCE_INFO();

    return ret;
}

static av_cold int vpe_vip_nbg_init(AVFilterContext *av_ctx)
{
    VPEVIPNBGContext *vip_nbg_ctx = av_ctx->priv;
    int i = 0;
    int ret = 0;

    /* process options and put options into dict */
    if(vip_nbg_ctx->vpe_vip_extra_params) {
        if((ret = av_dict_parse_string(&vip_nbg_ctx->dict, vip_nbg_ctx->vpe_vip_extra_params, "=", "|", 0)) < 0) {
            return ret;
        }
    }

    vip_nbg_ctx->vpe_vip_nbg_params.model_plugin_name = vip_nbg_ctx->osml_model;
    vip_nbg_ctx->vpe_vip_nbg_params.model_plugin_args = vip_nbg_ctx->osml_model_args;

    vip_nbg_ctx->vpe_vip_nbg_params.model_filename = "";

    av_log(av_ctx, AV_LOG_DEBUG, "get config JSON file name is %s\n",
                vip_nbg_ctx->vpe_vip_nbg_params.model_filename);

    vip_nbg_ctx->frames = av_calloc(vip_nbg_ctx->nb_inputs, sizeof(*vip_nbg_ctx->frames));
    if (!vip_nbg_ctx->frames)
        return AVERROR(ENOMEM);

    for(i = 0;i < vip_nbg_ctx->nb_inputs; i++){
        AVFilterPad in_pad = {0};

        in_pad.type = AVMEDIA_TYPE_VIDEO;
        in_pad.name = av_asprintf("input%d", i);
        if(!in_pad.name){
            av_free(vip_nbg_ctx->frames);
            for (i = 0; i < vip_nbg_ctx->nb_inputs;i++)
                av_freep(&av_ctx->input_pads[i].name);

            return AVERROR(ENOMEM);
        }

        if ((ret = ff_append_inpad(av_ctx, &in_pad)) < 0){
            av_free(vip_nbg_ctx->frames);
            av_freep(&in_pad.name);
            return ret;
        }
    }

    vip_nbg_ctx->vpe_vip_nbg_params.nb_frames = vip_nbg_ctx->nb_inputs;
    return 0;
}

static av_cold int init(AVFilterContext *ctx)
{
    int ret;
    VPEVIPNBGContext *s = ctx->priv;
    set_instance_name(s, ctx->name);

    {
        LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_ML, s->instance_name);
        ret = vpe_vip_nbg_init(ctx);
        LOG_AMA_RESTORE_INSTANCE_INFO();
    }
    return ret;
}

static av_cold void vpe_vip_nbg_uninit(AVFilterContext *ctx)
{
    VPEVIPNBGContext *vip_nbg_ctx = ctx->priv;
#if CONFIG_LIBXRM
    xrm_ml_release(&vip_nbg_ctx->xrm_ml_ctx);
#endif

    ff_framesync_uninit(&vip_nbg_ctx->fs);
    av_freep(&vip_nbg_ctx->frames);

    if (vip_nbg_ctx->in_hwframe_ctx)
        av_buffer_unref(&vip_nbg_ctx->in_hwframe_ctx);
    if (vip_nbg_ctx->out_hwframe_ctx)
        av_buffer_unref(&vip_nbg_ctx->out_hwframe_ctx);

    av_buffer_unref(&vip_nbg_ctx->hwdevice);

    if (vip_nbg_ctx->vpe_vip_nbg_handle)
        vpi_processor_close(vip_nbg_ctx->vpe_vip_nbg_handle);

    av_dict_free(&vip_nbg_ctx->dict);

    for (int i = 0; i < ctx->nb_inputs; i++)
        av_freep(&ctx->input_pads[i].name);
}

static av_cold void uninit(AVFilterContext *ctx)
{
    VPEVIPNBGContext *s = ctx->priv;

    LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_ML, s->instance_name);
    vpe_vip_nbg_uninit(ctx);
    LOG_AMA_RESTORE_INSTANCE_INFO();
}

static const AVFilterPad vpe_vip_nbg_outputs[] = {
    {
        .name          = "default",
        .type          = AVMEDIA_TYPE_VIDEO,
        .config_props  = config_params,
    },
};

AVFilter ff_vf_vip_nbg = {
    .name          = "ml_ama",
    .description   = NULL_IF_CONFIG_SMALL("Apply vip_nbg filter to the input."),
    .priv_size     = sizeof(VPEVIPNBGContext),
    .priv_class    = &ml_ama_class,
    .init          = init,
    .uninit        = uninit,
    .activate      = activate,
    FILTER_QUERY_FUNC(query_formats),
    FILTER_OUTPUTS(vpe_vip_nbg_outputs),
    .flags          = AVFILTER_FLAG_DYNAMIC_INPUTS | AVFILTER_FLAG_SLICE_THREADS,
    .flags_internal = FF_FILTER_FLAG_HWFRAME_AWARE,
};
