/*
 * VeriSilicon hardware context
 * Copyright (C) 2020 VeriSilicon Holdings Co., Ltd. <>
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#if defined(__linux__)
#include <unistd.h>
#include <sys/ioctl.h>
#endif

#include "pixdesc.h"
#include "hwcontext_internal.h"
#include "hwcontext_vpe.h"
#include "libavutil/opt.h"
#include "libavutil/thread.h"
#include "libavutil/buffer_internal.h"
#include "transfer_sidedata_vpe.h"
#if CONFIG_LIBXRM
#include <xrm_interface.h>
#endif

#include "log_ama.h"

typedef struct VpeFramesContext {
    VpiProcHandle hwdownload_handle;
    int hwdownload_init;
    VpiProcHandle hwupload_handle;
    int hwupload_init;
} VpeFramesContext;

static int vpe_get_vpiformat_from_avpixfmt(int format)
{
    switch (format) {
    case AV_PIX_FMT_NV12:
        return VPI_FORMAT_NV12;
    case AV_PIX_FMT_NV21:
        return VPI_FORMAT_NV21;
    case AV_PIX_FMT_P010LE:
        return VPI_FORMAT_P010LE;
    case AV_PIX_FMT_P010BE:
        return VPI_FORMAT_P010BE;
    case AV_PIX_FMT_YUV420P:
        return VPI_FORMAT_YUV420P;
    case AV_PIX_FMT_YUV420P10LE:
        return VPI_FORMAT_YUV420P10LE;
    case AV_PIX_FMT_YUV420P10BE:
        return VPI_FORMAT_YUV420P10BE;
    case AV_PIX_FMT_YUV422P:
        return VPI_FORMAT_YUV422P;
    case AV_PIX_FMT_YUV422P10LE:
        return VPI_FORMAT_YUV422P10LE;
    case AV_PIX_FMT_YUV422P10BE:
        return VPI_FORMAT_YUV422P10BE;
    case AV_PIX_FMT_YUV444P:
        return VPI_FORMAT_YUV444P;
    case AV_PIX_FMT_ARGB:
        return VPI_FORMAT_ARGB;
    case AV_PIX_FMT_ABGR:
        return VPI_FORMAT_ABGR;
    case AV_PIX_FMT_RGBA:
        return VPI_FORMAT_RGBA;
    case AV_PIX_FMT_BGRA:
        return VPI_FORMAT_BGRA;
    case AV_PIX_FMT_BGR0:
        return VPI_FORMAT_BGR0;
    case AV_PIX_FMT_RGB24:
        return VPI_FORMAT_RGB24;
    case AV_PIX_FMT_BGR24:
        return VPI_FORMAT_BGR24;
    case AV_PIX_FMT_RGB565LE:
        return VPI_FORMAT_RGB16;
    case AV_PIX_FMT_VPE:
        return VPI_FORMAT_VPE;
    case AV_PIX_FMT_RGBP:
        return VPI_FORMAT_RGB24_P;
    case AV_PIX_FMT_R16G16B16:
        return VPI_FORMAT_RGB48;
    case AV_PIX_FMT_R16G16B16P:
        return VPI_FORMAT_RGB48_P;
    case AV_PIX_FMT_BGRP:
        return VPI_FORMAT_BGR24_P;
    case AV_PIX_FMT_B16G16R16:
        return VPI_FORMAT_BGR48;
    case AV_PIX_FMT_B16G16R16P:
        return VPI_FORMAT_BGR48_P;
    case AV_PIX_FMT_PACKED10:
        return VPI_FORMAT_SP1010;
    default:
        av_log(NULL, AV_LOG_ERROR, "unsupport AVPixelFormat %d\n", format);
        return -1;
    }
}

static int vpe_get_avpixfmt_from_vpiformat(VpiFmt format)
{
    switch (format) {
    case VPI_FORMAT_NV12:
        return AV_PIX_FMT_NV12;
    case VPI_FORMAT_NV21:
        return AV_PIX_FMT_NV21;
    case VPI_FORMAT_P010LE:
        return AV_PIX_FMT_P010LE;
    case VPI_FORMAT_P010BE:
        return AV_PIX_FMT_P010BE;
    case VPI_FORMAT_YUV420P:
        return AV_PIX_FMT_YUV420P;
    case VPI_FORMAT_YUV420P10LE:
        return AV_PIX_FMT_YUV420P10LE;
    case VPI_FORMAT_YUV420P10BE:
        return AV_PIX_FMT_YUV420P10BE;
    case VPI_FORMAT_YUV422P:
        return AV_PIX_FMT_YUV422P;
    case VPI_FORMAT_YUV422P10LE:
        return AV_PIX_FMT_YUV422P10LE;
    case VPI_FORMAT_YUV422P10BE:
        return AV_PIX_FMT_YUV422P10BE;
    case VPI_FORMAT_YUV444P:
        return AV_PIX_FMT_YUV444P;
    case VPI_FORMAT_ARGB:
        return AV_PIX_FMT_ARGB;
    case VPI_FORMAT_ABGR:
        return AV_PIX_FMT_ABGR;
    case VPI_FORMAT_RGBA:
        return AV_PIX_FMT_RGBA;
    case VPI_FORMAT_BGRA:
        return AV_PIX_FMT_BGRA;
    case VPI_FORMAT_BGR0:
        return AV_PIX_FMT_BGR0;
    case VPI_FORMAT_RGB24:
        return AV_PIX_FMT_RGB24;
    case VPI_FORMAT_BGR24:
        return AV_PIX_FMT_BGR24;
    case VPI_FORMAT_RGB16:
        return AV_PIX_FMT_RGB565LE;
    case VPI_FORMAT_VPE:
        return AV_PIX_FMT_VPE;
    case VPI_FORMAT_RGB24_P:
        return AV_PIX_FMT_RGBP;
    case VPI_FORMAT_RGB48:
        return AV_PIX_FMT_R16G16B16;
    case VPI_FORMAT_RGB48_P:
        return AV_PIX_FMT_R16G16B16P;
    case VPI_FORMAT_BGR24_P:
        return AV_PIX_FMT_BGRP;
    case VPI_FORMAT_BGR48:
        return AV_PIX_FMT_B16G16R16;
    case VPI_FORMAT_BGR48_P:
        return AV_PIX_FMT_B16G16R16P;
    case VPI_FORMAT_SP1010:
        return AV_PIX_FMT_PACKED10;
    default:
        av_log(NULL, AV_LOG_ERROR, "unsupport VPI format %d\n", format);
        return -1;
    }
}

static int _vpe_frames_get_constraints(AVHWDeviceContext *ctx,
                                      const void *hwconfig,
                                      AVHWFramesConstraints *constraints)
{
    AVVpeDeviceContext *hwctx = ctx->hwctx;
    const AVVpeHWConfig *config = hwconfig;
    VpiHwCfg* hwcfg = NULL;
    int ret;
    int i;
    VpiFmt *pfmt;
    int sw_fmt_cnt;
    enum AVPixelFormat *sw_fmts;
    int dev_hwcfg = 0;

    av_log(ctx, AV_LOG_DEBUG, "dev_cfg.id = %d, config = %p, config->id = %d\n",
                 hwctx->dev_cfg->id, config, config ? config->id : 0);
    if (hwctx->dev_cfg->id && config && config->id != hwctx->dev_cfg->id) {
#ifdef SUPPORT_OSAL
        hwcfg = vpi_osal_get_hw_config(hwctx->vpe_dev_ctx, config->id);
#else
        hwcfg = vpi_device_get_hw_config(hwctx->vpe_dev_ctx, config->id);
#endif
        if (hwcfg == NULL) {
            return AVERROR_EXTERNAL;
        }

        sw_fmt_cnt = 0;
        pfmt = hwcfg->sw_formats;
        while (1) {
            if (*pfmt++ == VPI_FORMAT_NONE) {
                break;
            }
            sw_fmt_cnt++;
        }
        av_log(ctx, AV_LOG_DEBUG, "sw_fmt_cnt = %d\n", sw_fmt_cnt);

        sw_fmts = av_malloc_array(sw_fmt_cnt+1, sizeof(*sw_fmts));
        if (sw_fmts == NULL)
            return AVERROR(ENOMEM);
        av_log(ctx, AV_LOG_DEBUG, "sw_fmts = %p\n", sw_fmts);

        for (i = 0; i < sw_fmt_cnt; i++) {
            ret = vpe_get_avpixfmt_from_vpiformat(hwcfg->sw_formats[i]);
            av_log(ctx, AV_LOG_DEBUG, "%d -> %d\n", hwcfg->sw_formats[i], ret);
            if (ret < 0) {
                return AVERROR_EXTERNAL;
            }
            sw_fmts[i] = ret;
        }
        sw_fmts[sw_fmt_cnt] = AV_PIX_FMT_NONE;
        dev_hwcfg = 1;
    } else {
        sw_fmt_cnt = hwctx->sw_fmt_cnt;
        sw_fmts    = hwctx->supported_sw_formats;
        hwcfg      = hwctx->dev_cfg;
    }

    constraints->valid_sw_formats = av_malloc_array(sw_fmt_cnt+1, sizeof(*constraints->valid_sw_formats));
    if (!constraints->valid_sw_formats)
        return AVERROR(ENOMEM);

    for (i = 0; i < sw_fmt_cnt + 1; i++) {
        constraints->valid_sw_formats[i] = sw_fmts[i];
    }

    constraints->valid_hw_formats = av_malloc_array(hwctx->hw_fmt_cnt + 1,
                                                    sizeof(*constraints->valid_hw_formats));
    if (!constraints->valid_hw_formats)
        return AVERROR(ENOMEM);

    for (i = 0; i < hwctx->hw_fmt_cnt + 1; i++) {
        constraints->valid_hw_formats[i] = hwctx->supported_hw_formats[i];
    }

    av_log(ctx, AV_LOG_DEBUG, "[%s@%d], hwcfg.min_width %d, hwcfg.min_height %d, hwcfg.max_width %d, hwcfg.max_height %d\n",
            __FUNCTION__, __LINE__, hwcfg->min_width, hwcfg->min_height, hwcfg->max_width, hwcfg->max_height);
    constraints->min_width  = hwcfg->min_width;
    constraints->min_height = hwcfg->min_height;
    constraints->max_width  = hwcfg->max_width;
    constraints->max_height = hwcfg->max_height;

    if (sw_fmts != hwctx->supported_sw_formats)
        av_free(sw_fmts);

    if (hwcfg && dev_hwcfg) {
        vpi_device_free_hw_config(hwcfg);
    }

    return 0;
}

static int _vpe_transfer_get_formats(AVHWFramesContext *ctx,
                                    enum AVHWFrameTransferDirection dir,
                                    enum AVPixelFormat **formats)
{
    enum AVPixelFormat *fmts;
    AVVpeFramesContext *frame_hwctx = ctx->hwctx;

    fmts = av_malloc_array(3, sizeof(*fmts));
    if (!fmts)
        return AVERROR(ENOMEM);

    if (dir == AV_HWFRAME_TRANSFER_DIRECTION_FROM) {
        fmts[0] = frame_hwctx->sw_format[0];
        fmts[1] = frame_hwctx->sw_format[1];
        fmts[2] = AV_PIX_FMT_NONE;
    } else {
        fmts[0] = AV_PIX_FMT_VPE;
        fmts[1] = AV_PIX_FMT_NONE;
    }

    *formats = fmts;

    return 0;
}

static int vpe_init_internal_session(AVHWFramesContext *ctx, int download)
{
    AVVpeDeviceContext *device_hwctx = ctx->device_ctx->hwctx;
    VpeFramesContext *priv           = ctx->internal->priv;
    AVVpeFramesContext *frame_hwctx  = ctx->hwctx;
    VpiProcInitParams *param;
    int ret;

    if (!priv->hwdownload_handle && !priv->hwdownload_init && download) {
        //Create HWDOWNLOAD_VPE
        param = (VpiProcInitParams *)vpi_memory_get_struct(VPE_PROC_INIT_PARAM);
        if (!param) {
            return AVERROR(ENOMEM);
        }
        param->processor_type = VPE_PROCESSOR_HWDOWNLOAD;
        vpi_device_set_ctx(param, device_hwctx->vpe_dev_ctx, VPE_PROCESSOR_HWDOWNLOAD);

        param->mem_ctx = frame_hwctx->vpe_mem_ctx;

        ret = vpi_processor_init(param, &priv->hwdownload_handle);
        if (ret < 0){
            av_log(ctx, AV_LOG_ERROR, "VPE Hwdownload init error: %s\n", vpi_error_str(ret));
            return AVERROR(ENOMEM);
        }

        priv->hwdownload_init = 1;
        vpi_memory_free_struct(param);
    }

    if (!priv->hwupload_handle && !priv->hwupload_init && !download) {
        //Create HWUPLOAD_VPE
        param = (VpiProcInitParams *)vpi_memory_get_struct(VPE_PROC_INIT_PARAM);
        if (!param) {
            return AVERROR(ENOMEM);
        }
        param->processor_type = VPE_PROCESSOR_HWUPLOAD;
        vpi_device_set_ctx(param, device_hwctx->vpe_dev_ctx, VPE_HW_UPLOAD);
        param->mem_ctx = frame_hwctx->vpe_mem_ctx;
        param->width = ctx->width;
        param->height = ctx->height;
        param->input_sw_format = av_get_pix_fmt_name(ctx->sw_format);
        ret = vpi_processor_init(param, &priv->hwupload_handle);
        if (ret < 0){
            av_log(ctx, AV_LOG_ERROR, "VPE Hwupload init error: %s\n", vpi_error_str(ret));
            return AVERROR(ENOMEM);
        }

        av_log(ctx, AV_LOG_DEBUG, "input sw format %s\n", param->input_sw_format);
        priv->hwupload_init = 1;
        vpi_memory_free_struct(param);
    }
    return 0;
}

static int _vpe_transfer_data_from(AVHWFramesContext *ctx, AVFrame *dst,
                                  const AVFrame *src)
{
    AVVpeFramesContext *fmctx = ctx->hwctx;
    VpeFramesContext *priv = ctx->internal->priv;
    VpiFrm *in_frame;
    VpiFrm *out_frame;
    int pp_index = -1;
    int in_ch;
    int plane;
    int ret;

    ret = vpe_init_internal_session(ctx, 1);
    if (ret < 0) {
        return AVERROR(ENOMEM);
    }

    ret = vpi_frame_alloc(fmctx->vpe_mem_ctx, &out_frame);
    if (ret < 0) {
        return ret;
    }

    ret = vpi_memory_get_channel_info(fmctx->vpe_mem_ctx, &fmctx->proc_info->ch_info[0]);
    if (ret < 0) {
        av_log(ctx, AV_LOG_ERROR, "vpi_memory_get_info failed\n");
        return AVERROR_EXTERNAL;
    }

    if (src->buf[1]) {
        in_ch = *(int *)(src->buf[1]->data);
    } else {
        in_ch = 0;
    }
    ret = vpi_memory_get_map_info(fmctx->vpe_mem_ctx, in_ch, &pp_index);
    if (ret < 0) {
        av_log(ctx, AV_LOG_ERROR, "vpi_memory_get_map_info failed\n");
        return AVERROR_EXTERNAL;
    }

    in_frame = (VpiFrm *)src->data[0];
    out_frame->width  = in_frame->ch_info[pp_index].width;
    out_frame->height = in_frame->ch_info[pp_index].height;

    out_frame->ch_index = in_ch;
    out_frame->ch_info[in_ch].enabled   = 1;
    out_frame->ch_info[in_ch].sw_format = in_frame->ch_info[pp_index].sw_format;
    out_frame->ch_info[in_ch].width     = out_frame->width;
    out_frame->ch_info[in_ch].height    = out_frame->height;

    for (plane = 0; plane < av_pix_fmt_count_planes(ctx->sw_format); plane++) {
        if (dst->data[plane]) {
            av_log(ctx, AV_LOG_DEBUG, "dst->data[%d] = %p\n", plane, dst->data[plane]);
            av_log(ctx, AV_LOG_DEBUG, "dst->linesize[%d] = %d\n", plane, dst->linesize[plane]);
            ret = vpi_pic_set_buffer(out_frame, in_ch, plane, dst->data[plane], dst->linesize[plane]);
            if (ret < 0) {
                return AVERROR(ENOMEM);
            }
        }
    }

    vpe_transfer_side_data_from((VpiFrm *)in_frame, dst);

    ret = vpi_processor_process_frame(priv->hwdownload_handle, in_frame, out_frame);
    if (ret < 0) {
        av_log(ctx, AV_LOG_ERROR, "vpi_processor_process_frame failed, error=%s(%d)\n", vpi_error_str(ret), ret);
        return AVERROR_EXTERNAL;
    }

    dst->key_frame = src->key_frame;

    vpi_frame_free(fmctx->vpe_mem_ctx, out_frame);

    return 0;
}

static int _vpe_transfer_data_to(AVHWFramesContext *ctx, AVFrame *dst,
                                const AVFrame *src)
{
    AVVpeFramesContext *fmctx = ctx->hwctx;
    VpeFramesContext *priv    = ctx->internal->priv;
    VpiFrm *in_frame;
    VpiFrm *out_frame;
    int ret;
    int plane;

    av_log(ctx, AV_LOG_DEBUG, "input sw format %d, %s\n", src->format, av_get_pix_fmt_name(src->format));

    ret = vpe_init_internal_session(ctx, 0);
    if (ret < 0) {
        return ret;
    }

    ret = vpi_frame_alloc(fmctx->vpe_mem_ctx, &in_frame);
    if (ret < 0) {
        return AVERROR(ENOMEM);
    }

    av_log(ctx, AV_LOG_DEBUG, "input width %d, height %d\n", src->width, src->height);
    av_log(ctx, AV_LOG_DEBUG, "linesize %d, %d, %d\n", src->linesize[0], src->linesize[1], src->linesize[2]);

    in_frame->width  = src->width;
    in_frame->height = src->height;
    in_frame->format = vpe_get_vpiformat_from_avpixfmt(src->format);
    in_frame->ch_info[0].enabled   = 1;
    in_frame->ch_info[0].format    = vpe_get_vpiformat_from_avpixfmt(src->format);
    in_frame->ch_info[0].sw_format = vpe_get_vpiformat_from_avpixfmt(src->format);
    in_frame->ch_info[0].width     = in_frame->width;
    in_frame->ch_info[0].height    = in_frame->height;
    in_frame->pts = src->pts;
    in_frame->dts = src->pkt_dts;
    in_frame->key_frame = src->key_frame;
    in_frame->poc = src->coded_picture_number;

    for (plane = 0; plane < av_pix_fmt_count_planes(ctx->sw_format); plane++) {
        if (src->data[plane]) {
            av_log(ctx, AV_LOG_DEBUG, "src->data[%d] = %p\n", plane, src->data[plane]);
            av_log(ctx, AV_LOG_DEBUG, "src->linesize[%d] = %d\n", plane, src->linesize[plane]);
            ret = vpi_pic_set_buffer(in_frame, 0, plane, src->data[plane], src->linesize[plane]);
            if (ret < 0) {
                return AVERROR(ENOMEM);
            }
        }
    }

    out_frame = (VpiFrm *)dst->data[0];

    vpe_transfer_side_data_to(priv->hwupload_handle, src, out_frame);

    ret = vpi_processor_process_frame(priv->hwupload_handle, in_frame, out_frame);
    if (ret < 0) {
        av_log(ctx, AV_LOG_ERROR, "vpi_processor_process_frame failed, error=%s(%d)\n", vpi_error_str(ret), ret);
        return AVERROR_EXTERNAL;
    }

    vpi_frame_free(fmctx->vpe_mem_ctx, in_frame);

    return 0;
}

static void vpe_buffer_free(void *opaque, uint8_t *data)
{
    AVHWFramesContext  *ctx    = opaque;
    AVVpeFramesContext *fmctx  = ctx->hwctx;
    av_log(ctx, AV_LOG_DEBUG, "%s(%d) %p\n", __FUNCTION__, __LINE__, data);

    vpi_frame_free(fmctx->vpe_mem_ctx, (VpiFrm *)data);
}

static AVBufferRef *vpe_pool_alloc(void *opaque, size_t size)
{
    AVHWFramesContext  *ctx    = opaque;
    AVVpeFramesContext *fmctx  = ctx->hwctx;
    AVBufferRef *buf;
    VpiFrm *v_frame;
    int ret;

    av_log(ctx, AV_LOG_DEBUG, "%s(%d) size = %d\n", __FUNCTION__, __LINE__, (int)size);

    ret = vpi_frame_alloc(fmctx->vpe_mem_ctx, &v_frame);
    if (ret < 0)
        return NULL;

    buf = av_buffer_create((uint8_t*)v_frame, (int)size,
                            vpe_buffer_free, ctx, 0);
    if (!buf) {
        vpi_frame_free(fmctx->vpe_mem_ctx, v_frame);
        return buf;
    }

    av_log(ctx, AV_LOG_DEBUG, "%s(%d) pic = %p\n", __FUNCTION__, __LINE__, v_frame);

    v_frame->width  = ctx->width;
    v_frame->height = ctx->height;
    v_frame->format = VPI_FORMAT_VPE;
    av_log(ctx, AV_LOG_DEBUG, "%s(%d) buf = %p\n", __FUNCTION__, __LINE__, buf);
    av_log(ctx, AV_LOG_DEBUG, "%s(%d) buf->data = %p\n", __FUNCTION__, __LINE__, buf->data);

    return buf;
}

static int _vpe_frames_init(AVHWFramesContext *hwfc)
{
    AVVpeDeviceContext *hwctx = hwfc->device_ctx->hwctx;
    AVVpeFramesContext *fmctx = hwfc->hwctx;
    VpeFramesContext *priv    = hwfc->internal->priv;
    enum AVPixelFormat *pfmt;
    int i;

    av_log(hwfc, AV_LOG_DEBUG, "hwfc is %p\n", hwfc);
    av_log(hwfc, AV_LOG_DEBUG, "hwfc->sw_format is %s\n", av_get_pix_fmt_name(hwfc->sw_format));
    av_log(hwfc, AV_LOG_DEBUG, "hwfc->width = %d, hwfc->height = %d\n", hwfc->width, hwfc->height);

    if (hwctx->supported_sw_formats == NULL) {
        av_log(hwfc, AV_LOG_ERROR, "hwctx supported_sw_formats NULL\n");
        return AVERROR_UNKNOWN;
    }

    pfmt = hwctx->supported_sw_formats;
    while (1) {
        if (hwfc->sw_format == *pfmt)
            break;
        pfmt++;
    }
    if (*pfmt == AV_PIX_FMT_NONE) {
        av_log(hwfc, AV_LOG_ERROR, "Pixel format '%s' is not supported\n",
               av_get_pix_fmt_name(hwfc->sw_format));
        return AVERROR(ENOSYS);
    }

#ifdef SUPPORT_OSAL
    fmctx->vpe_mem_ctx = vpi_osal_get_memctx(hwctx->vpe_dev_ctx);
#else
    fmctx->vpe_mem_ctx = vpi_memory_init(hwctx->vpe_handle, hwctx->vpe_dev_ctx);
#endif
    if (fmctx->vpe_mem_ctx == NULL) {
        return AVERROR_EXTERNAL;
    }

    if (!hwfc->pool) {
        hwfc->internal->pool_internal =
            av_buffer_pool_init2(sizeof(VpiFrm), hwfc, vpe_pool_alloc, NULL);
        if (!hwfc->internal->pool_internal) {
            av_log(hwfc, AV_LOG_ERROR, "%s(%d)\n", __FUNCTION__, __LINE__);
            return AVERROR(ENOMEM);
        }
    }

    fmctx->proc_info = vpi_memory_get_struct(VPE_PROC_INFO);
    if (!fmctx->proc_info) {
        return AVERROR(ENOMEM);
    }

    // for support rgb/yuv
    fmctx->sw_format = av_malloc_array(3, sizeof(*fmctx->sw_format));
    if (fmctx->sw_format == NULL)
        return AVERROR(ENOMEM);
    fmctx->sw_format[0] = hwfc->sw_format;
    for (i = 1; i < 3; i++)
        fmctx->sw_format[i] = AV_PIX_FMT_NONE;

    vpi_memory_get_channel_info(fmctx->vpe_mem_ctx, &fmctx->proc_info->ch_info[0]);

    // set the init map info
    if (vpi_memory_set_map_info(fmctx->vpe_mem_ctx, 0, 0)) {
        av_log(hwfc, AV_LOG_ERROR, "init vpi_memory_set_map_info failed\n");
        return AVERROR_EXTERNAL;
    }

    priv->hwdownload_handle = NULL;
    priv->hwdownload_init   = 0;

    priv->hwupload_handle = NULL;
    priv->hwupload_init   = 0;

    return 0;
}

static void _vpe_frames_uninit(AVHWFramesContext *ctx)
{
    AVVpeFramesContext *fmctx = ctx->hwctx;
    VpeFramesContext *priv = ctx->internal->priv;

    if (fmctx->proc_info) {
        vpi_memory_free_struct(fmctx->proc_info);
    }

    if (priv->hwdownload_handle) {
        vpi_processor_close(priv->hwdownload_handle);
        priv->hwdownload_handle = NULL;
        priv->hwdownload_init   = 0;
    }
    if (priv->hwupload_handle) {
        vpi_processor_close(priv->hwupload_handle);
        priv->hwupload_handle = NULL;
        priv->hwupload_init   = 0;
    }

    av_free(fmctx->sw_format);
    vpi_memory_release(fmctx->vpe_mem_ctx);
}

static void vpe_free_buffer(void *opaque, uint8_t *data)
{
    VpiFrm *vpi_frame = (VpiFrm *)data;

    if (vpi_frame->buf_ref[0])
        av_frame_unref((AVFrame *)vpi_frame->buf_ref[0]);

    //vpi_frame_release(vpi_frame);
    av_buffer_pool_release_buffer(opaque, data);
}

static int _vpe_get_buffer(AVHWFramesContext *hwfc, AVFrame *frame)
{
    av_log(hwfc, AV_LOG_DEBUG, "%s(%d) %p\n", __FUNCTION__, __LINE__, hwfc->pool);

    frame->buf[0] = av_buffer_pool_get(hwfc->pool);
    if (!frame->buf[0]) {
        av_log(hwfc, AV_LOG_ERROR, "%s(%d)\n", __FUNCTION__, __LINE__);
        return AVERROR(ENOMEM);
    }
    frame->data[0] = frame->buf[0]->data;
    frame->buf[0]->buffer->free = vpe_free_buffer;
    frame->format = hwfc->format;
    frame->width  = hwfc->width;
    frame->height = hwfc->height;

    //Reset the sidedata pointers
    VpiFrm* vpi_frame;
    vpi_frame = (VpiFrm*)frame->data[0];
    memset(vpi_frame, 0, sizeof(VpiFrm));
    //memset(vpi_frame->vpi_sidedata, 0, sizeof(vpi_frame->vpi_sidedata));
    return 0;
}

static void _vpe_device_uninit(AVHWDeviceContext *device_ctx)
{
    AVVpeDeviceContext *hwctx = device_ctx->hwctx;
    av_log(device_ctx, AV_LOG_DEBUG, "%s(%d)\n", __FUNCTION__, __LINE__);

    av_freep(&hwctx->supported_hw_formats);
    av_freep(&hwctx->supported_sw_formats);
#if !defined(SUPPORT_OSAL)
    if (hwctx->dev_cfg) {
        vpi_device_free_hw_config(hwctx->dev_cfg);
        hwctx->dev_cfg = NULL;
    }
#endif
}

static int _vpe_device_init(AVHWDeviceContext *device_ctx)
{
    AVVpeDeviceContext *hwctx = device_ctx->hwctx;
    VpiFmt *vpi_fmt;
    int ret;
    int i;

    av_log(device_ctx, AV_LOG_DEBUG, "%s(%d)\n", __FUNCTION__, __LINE__);
#ifdef SUPPORT_OSAL
    hwctx->dev_cfg = vpi_osal_get_hw_config(hwctx->vpe_dev_ctx, VPE_HW_COMMON);
#else
    hwctx->dev_cfg = vpi_device_get_hw_config(hwctx->vpe_dev_ctx, VPE_HW_COMMON);
#endif
    if (hwctx->dev_cfg == NULL) {
        av_log(device_ctx, AV_LOG_ERROR, "vpi_device_get_hw_config failed\n");
        return AVERROR_EXTERNAL;
    }

    hwctx->sw_fmt_cnt = 0;
    vpi_fmt = hwctx->dev_cfg->sw_formats;
    while (1) {
        if (*vpi_fmt++ == VPI_FORMAT_NONE) {
            break;
        }
        hwctx->sw_fmt_cnt++;
    }
    av_log(device_ctx, AV_LOG_DEBUG, "hwctx->sw_fmt_cnt = %d\n", hwctx->sw_fmt_cnt);

    hwctx->supported_sw_formats = av_malloc_array(hwctx->sw_fmt_cnt+1, sizeof(*hwctx->supported_sw_formats));
    if (hwctx->supported_sw_formats == NULL)
        return AVERROR(ENOMEM);
    av_log(device_ctx, AV_LOG_DEBUG, "supported_sw_formats = %p\n", hwctx->supported_sw_formats);

    for (i = 0; i < hwctx->sw_fmt_cnt; i++) {
        ret = vpe_get_avpixfmt_from_vpiformat(hwctx->dev_cfg->sw_formats[i]);
        av_log(device_ctx, AV_LOG_DEBUG, "%d -> %d\n", hwctx->dev_cfg->sw_formats[i], ret);
        if (ret < 0) {
            return AVERROR_EXTERNAL;
        }
        hwctx->supported_sw_formats[i] = ret;
    }
    hwctx->supported_sw_formats[hwctx->sw_fmt_cnt] = AV_PIX_FMT_NONE;

    hwctx->hw_fmt_cnt = 0;
    vpi_fmt = hwctx->dev_cfg->hw_formats;
    while (1) {
        if (*vpi_fmt++ == VPI_FORMAT_NONE) {
            break;
        }
        hwctx->hw_fmt_cnt++;
    }
    av_log(device_ctx, AV_LOG_DEBUG, "hwctx->hw_fmt_cnt = %d\n", hwctx->hw_fmt_cnt);

    hwctx->supported_hw_formats = av_malloc_array(hwctx->hw_fmt_cnt+1, sizeof(*hwctx->supported_hw_formats));
    if (hwctx->supported_hw_formats == NULL)
        return AVERROR(ENOMEM);
    av_log(device_ctx, AV_LOG_DEBUG, "supported_hw_formats = %p\n", hwctx->supported_hw_formats);

    for (i = 0; i < hwctx->hw_fmt_cnt; i++) {
        ret = vpe_get_avpixfmt_from_vpiformat(hwctx->dev_cfg->hw_formats[i]);
        av_log(device_ctx, AV_LOG_DEBUG, "%d -> %d\n", hwctx->dev_cfg->hw_formats[i], ret);
        if (ret < 0) {
            return AVERROR_EXTERNAL;
        }
        hwctx->supported_hw_formats[i] = ret;
    }
    hwctx->supported_hw_formats[hwctx->hw_fmt_cnt] = AV_PIX_FMT_NONE;

  	return 0;
}

static void vpe_device_free(AVHWDeviceContext *device_ctx)
{
    AVVpeDeviceContext *hwctx = device_ctx->hwctx;

#ifdef SUPPORT_OSAL
    if (hwctx->vpe_dev_ctx) {
        vpi_osal_destroy_hwcontext(hwctx->vpe_dev_ctx);
    }
#else
    if (hwctx->vpe_dev_ctx) {
        vpi_device_close(hwctx->vpe_dev_ctx);
        vpi_device_destroy(hwctx->vpe_dev_ctx);
        hwctx->vpe_dev_ctx = NULL;
    }

    if (hwctx->vpe_handle) {
        vpi_destroy(hwctx->vpe_handle);
        hwctx->vpe_handle = NULL;
    }
#endif
}

static char* vpe_device_get_device(AVHWDeviceContext *device_ctx, const char* p_device)
{
    int dev_index = 0;
    int amount_written = 0;
    int device_len = p_device == NULL ? strlen("/dev/ama_transcoder999") + 1 : strlen(p_device) + 1;
    char* device = calloc(1, device_len);
    if (!device) {
        av_log(device_ctx, AV_LOG_ERROR, "Unable to allocate memory for device location!\n");
        return NULL;
    }

    if (p_device && getenv("XRM_RESERVE_ID")) {
        free(device);
        av_log(device_ctx, AV_LOG_ERROR, "Both device specified and xrm reservation ID is set!\n");
        return NULL;
    }
#if CONFIG_LIBXRM
    dev_index = xrm_interface_get_dev_index();
    if (dev_index < 0 || dev_index > 999) { // Accommodate up to 3 digits
        free(device);
        return NULL;
    }
#endif
    if (p_device) {
        amount_written = snprintf(device, strlen(p_device) + 1, "%s", p_device);
    } else {
        amount_written = snprintf(device, device_len, "/dev/ama_transcoder%d", dev_index);
    }
    if (amount_written <= 0 || amount_written > device_len) {
        free(device);
        av_log(device_ctx, AV_LOG_ERROR, "Unable to set device location!\n");
        return NULL;
    }
    return device;
}

static int _vpe_device_create(AVHWDeviceContext *device_ctx, const char *p_device,
                             AVDictionary *opts, int flags)
{
    AVVpeDeviceContext *hwctx = device_ctx->hwctx;
    AVDictionaryEntry *opt = NULL;
    char* device = vpe_device_get_device(device_ctx, p_device);
    int ret = 0;

    if (!device) {
        return AVERROR_EXTERNAL;
    }

    device_ctx->free = vpe_device_free;

    hwctx->vpe_dev_ctx = vpi_osal_create_hwcontext(device);
    if (hwctx->vpe_dev_ctx == NULL) {
        av_log(device_ctx, AV_LOG_ERROR, "osl create hwcontext failed\n");
        return AVERROR_EXTERNAL;
    }
    free(device);

    opt = NULL;
    while (opt = av_dict_get(opts, "", opt, AV_DICT_IGNORE_SUFFIX)) {
        printf("opt->key %s, opt->value %s\n", opt->key, opt->value);
#ifdef SUPPORT_OSAL
        if (vpi_osal_set_param(hwctx->vpe_dev_ctx, opt->key, opt->value) < 0)
#else
        if (vpi_device_set_param(hwctx->vpe_dev_ctx, opt->key, opt->value) < 0)
#endif
            av_log(device_ctx, AV_LOG_WARNING,
                   "Error parsing option '%s = %s'. Please check if it is valid for vpe device extra options.\n",
                   opt->key, opt->value);
    }
#ifndef SUPPORT_OSAL
    if (vpi_device_open(hwctx->vpe_dev_ctx))
        return AVERROR_EXTERNAL;
    }
#endif
    return 0;
}

static int vpe_device_create(AVHWDeviceContext *device_ctx, const char *p_device,
                             AVDictionary *opts, int flags)
{
    int ret;
    LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_GENERAL, "ama_hwcontext");
    ret = _vpe_device_create(device_ctx, p_device, opts, flags);
    LOG_AMA_RESTORE_INSTANCE_INFO();
    return ret;
}

static int vpe_device_init(AVHWDeviceContext *device_ctx)
{
    int ret;
    LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_GENERAL, "ama_hwcontext");
    ret = _vpe_device_init(device_ctx);
    LOG_AMA_RESTORE_INSTANCE_INFO();
    return ret;
}

static void vpe_device_uninit(AVHWDeviceContext *device_ctx)
{
    LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_GENERAL, "ama_hwcontext");
    _vpe_device_uninit(device_ctx);
    LOG_AMA_RESTORE_INSTANCE_INFO();
}

static int vpe_frames_get_constraints(AVHWDeviceContext *ctx,
                                      const void *hwconfig,
                                      AVHWFramesConstraints *constraints)
{
    int ret;
    LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_GENERAL, "ama_hwcontext");
    ret = _vpe_frames_get_constraints(ctx, hwconfig, constraints);
    LOG_AMA_RESTORE_INSTANCE_INFO();
    return ret;
}

static int vpe_frames_init(AVHWFramesContext *hwfc)
{
    int ret;
    LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_GENERAL, "ama_hwcontext");
    ret = _vpe_frames_init(hwfc);
    LOG_AMA_RESTORE_INSTANCE_INFO();
    return ret;
}

static void vpe_frames_uninit(AVHWFramesContext *ctx)
{
    LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_GENERAL, "ama_hwcontext");
    _vpe_frames_uninit(ctx);
    LOG_AMA_RESTORE_INSTANCE_INFO();
}

static int vpe_get_buffer(AVHWFramesContext *hwfc, AVFrame *frame)
{
    int ret;
    LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_GENERAL, "ama_hwcontext");
    ret = _vpe_get_buffer(hwfc, frame);
    LOG_AMA_RESTORE_INSTANCE_INFO();
    return ret;
}

static int vpe_transfer_get_formats(AVHWFramesContext *ctx,
                                    enum AVHWFrameTransferDirection dir,
                                    enum AVPixelFormat **formats)
{
    int ret;
    LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_HDMA, "ama_hwcontext");
    ret = _vpe_transfer_get_formats(ctx, dir, formats);
    LOG_AMA_RESTORE_INSTANCE_INFO();
    return ret;
}

static int vpe_transfer_data_from(AVHWFramesContext *ctx, AVFrame *dst,
                                  const AVFrame *src)
{
    int ret;
    ret = _vpe_transfer_data_from(ctx, dst, src);
    return ret;
}

static int vpe_transfer_data_to(AVHWFramesContext *ctx, AVFrame *dst,
                                const AVFrame *src)
{
    int ret;
    ret = _vpe_transfer_data_to(ctx, dst, src);
    return ret;
}

const HWContextType ff_hwcontext_type_vpe = {
    .type = AV_HWDEVICE_TYPE_AMA,
    .name = "AMA",

    .device_hwctx_size = sizeof(AVVpeDeviceContext),
    .frames_hwctx_size = sizeof(AVVpeFramesContext),
    .frames_priv_size  = sizeof(VpeFramesContext),
    .device_hwconfig_size   = sizeof(AVVpeHWConfig),

    .device_create          = vpe_device_create,
    .device_init            = vpe_device_init,
    .device_uninit          = vpe_device_uninit,
    .frames_get_constraints = vpe_frames_get_constraints,
    .frames_init            = vpe_frames_init,
    .frames_uninit          = vpe_frames_uninit,
    .frames_get_buffer      = vpe_get_buffer,
    .transfer_get_formats   = vpe_transfer_get_formats,
    .transfer_data_from     = vpe_transfer_data_from,
    .transfer_data_to       = vpe_transfer_data_to,

    .pix_fmts = (const enum AVPixelFormat[]){ AV_PIX_FMT_VPE, AV_PIX_FMT_NONE },
};
